from typing import (
    Union,
    MutableMapping,
    Iterable,
    Tuple,
    Any,
    ClassVar,
    Optional,
    TypeVar,
    Generic,
    Hashable,
    Callable,
)
from numbers import Number
from copy import deepcopy

import numpy as np
from numpy import ndarray
from numpy.lib.index_tricks import IndexExpression

from sigmaepsilon.core.typing import issequence
from sigmaepsilon.math import atleast1d, atleast2d, atleastnd, ascont
from sigmaepsilon.math.linalg import ReferenceFrame as FrameLike
from sigmaepsilon.math.utils import to_range_1d
from sigmaepsilon.math.linalg.sparse import csr_matrix

from ..typing import (
    ABC_PolyCell,
    PolyDataProtocol,
    PointDataProtocol,
    GeometryProtocol,
    CellDataProtocol,
)
from .celldata import CellData
from ..space import PointCloud, CartesianFrame
from ..utils import (
    distribute_nodal_data_bulk,
    distribute_nodal_data_sparse,
)
from ..utils.utils import (
    jacobian_matrix_bulk,
    jacobian_matrix_bulk_1d,
    jacobian_det_bulk_1d,
    points_of_cells,
    pcoords_to_coords,
    pcoords_to_coords_1d,
    cells_coords,
    lengths_of_lines,
)
from ..utils.cells.utils import (
    _loc_to_glob_bulk_,
)
from ..utils.tet import (
    _pip_tet_bulk_knn_,
    _pip_tet_bulk_,
    _glob_to_nat_tet_bulk_,
    _glob_to_nat_tet_bulk_knn_,
    _pip_tet_bulk_nat_,
)
from ..utils.cells.utils import (
    _find_first_hits_,
    _find_first_hits_knn_,
    _ntet_to_loc_bulk_,
    cell_measures,
)
from ..utils.topology.topo import detach_mesh_bulk, rewire
from ..utils.topology import transform_topology
from ..utils.tri import _pip_tri_bulk_
from ..utils.knn import k_nearest_neighbours
from ..utils.space import index_of_closest_point, frames_of_lines, frames_of_surfaces
from ..utils import cell_centers_bulk
from ..vtkutils import mesh_to_UnstructuredGrid as mesh_to_vtk
from ..topoarray import TopologyArray
from ..space import CartesianFrame
from ..utils.numint import Quadrature
from ..config import __haspyvista__

if __haspyvista__:
    import pyvista as pv

MapLike = Union[ndarray, MutableMapping]
PointDataLike = TypeVar("PointDataLike", bound=PointDataProtocol)
MeshDataLike = TypeVar("MeshDataLike", bound=PolyDataProtocol)
T = TypeVar("T", bound="PolyCell")

__all__ = ["PolyCell"]


class PolyCell(Generic[MeshDataLike, PointDataLike], ABC_PolyCell):
    """
    A subclass of :class:`~sigmaepsilon.mesh.data.celldata.CellData` as a base class
    for all cell containers. The class should not be used directly, the main purpose
    here is encapsulation of common behaviour for all kinds of cells.
    """

    label: ClassVar[Optional[str]] = None
    Geometry: ClassVar[GeometryProtocol]

    data_class: ClassVar[type] = CellData[MeshDataLike, PointDataLike]

    def __init__(
        self,
        *args,
        db: Optional[Union[CellData[MeshDataLike, PointDataLike], None]] = None,
        pointdata: Optional[Union[PointDataLike, None]] = None,
        container: Optional[Union[MeshDataLike, None]] = None,
        **kwargs,
    ):
        if db is None:
            db_class = self.__class__.data_class
            db = db_class(*args, **kwargs)

        self._db = db
        self._pointdata = pointdata
        self._container = container

        super().__init__()

    @property
    def db(self) -> CellDataProtocol[MeshDataLike, PointDataLike]:
        """
        Returns the database of the block.
        """
        return self._db

    @db.setter
    def db(self, value: CellDataProtocol[MeshDataLike, PointDataLike]) -> None:
        """
        Sets the database of the block.
        """
        self._db = value

    @property
    def pointdata(self) -> PointDataLike:
        """
        Returns the hosting point database. This is what
        the topology of the cells are referring to.
        """
        return self._pointdata

    @pointdata.setter
    def pointdata(self, value: PointDataLike) -> None:
        """
        Sets the hosting point database. This is what
        the topology of the cells are referring to.
        """
        if value is not None:
            if not isinstance(value, PointDataProtocol):
                raise TypeError("'value' must be a PointData instance")
        self._pointdata = value

    @property
    def pd(self) -> PointDataLike:
        """
        Returns the attached point database. This is what
        the topology of the cells are referring to.
        """
        return self.pointdata

    @pd.setter
    def pd(self, value: PointDataLike) -> None:
        """
        Sets the attached pointdata.
        """
        self.pointdata = value

    @property
    def container(self) -> MeshDataLike:
        """
        Returns the container of the block.
        """
        return self._container

    @container.setter
    def container(self, value: MeshDataLike) -> None:
        """
        Sets the container of the block.
        """
        if not isinstance(value, PolyDataProtocol):
            raise TypeError("'value' must be a PolyData instance")
        self._container = value

    def root(self) -> MeshDataLike:
        """
        Returns the top level container of the model the block is
        the part of.
        """
        c = self.container
        return None if c is None else c.root

    def source(self) -> MeshDataLike:
        """
        Retruns the source of the cells. This is the PolyData block
        that stores the PointData object the topology of the cells
        are referring to.
        """
        c = self.container
        return None if c is None else c.source()

    def pull(
        self, data: Union[str, ndarray], ndf: Union[ndarray, csr_matrix] = None
    ) -> ndarray:
        """
        Pulls data from the attached pointdata. The pulled data is either copied or
        distributed according to a measure.

        Parameters
        ----------
        data: str or numpy.ndarray
            Either a field key to identify data in the database of the attached
            PointData, or a NumPy array.

        See Also
        --------
        :func:`~sigmaepsilon.mesh.utils.utils.distribute_nodal_data_bulk`
        :func:`~sigmaepsilon.mesh.utils.utils.distribute_nodal_data_sparse`
        """
        if isinstance(data, str):
            pd = self.source().pd
            nodal_data = pd[data].to_numpy()
        else:
            assert isinstance(
                data, ndarray
            ), "'data' must be a string or a NumPy array."
            nodal_data = data
        topo = self.nodes
        if ndf is None:
            ndf = np.ones_like(topo).astype(float)
        if len(nodal_data.shape) == 1:
            nodal_data = atleast2d(nodal_data, back=True)
        if isinstance(ndf, ndarray):
            d = distribute_nodal_data_bulk(nodal_data, topo, ndf)
        else:
            d = distribute_nodal_data_sparse(nodal_data, topo, self.id, ndf)
        # nE, nNE, nDATA
        return d

    def _get_cell_slicer(
        self, cells: Optional[Union[int, Iterable[int]]] = None
    ) -> Union[Iterable[int], IndexExpression]:
        if isinstance(cells, Iterable):
            cells = atleast1d(cells)
            conds = np.isin(cells, self.db.id)
            cells = atleast1d(cells[conds])
            assert (
                len(cells) > 0
            ), "Length of cells is zero. At least one cell must be requested"
        else:
            cells = np.s_[:]
        return cells

    def _get_points_and_range(
        self,
        points: Optional[Union[None, Iterable[Number]]] = None,
        rng: Optional[Union[None, Iterable[Number]]] = None,
    ) -> Tuple[ndarray, ndarray]:
        """
        Returns the points and the domain in which they are defined.
        Currently the range is only a matter of interest for 1d cells. Then,
        the points are returned in the range [-1, 1]. In the future it is planned
        to make the target range a parameter of the function.
        """
        nDIM = self.Geometry.number_of_spatial_dimensions
        if nDIM == 1:
            if points is None:
                points = np.array(self.Geometry.master_coordinates()).flatten()
                rng = [-1, 1]
            else:
                points = atleast1d(np.array(points))
                rng = np.array([-1, 1]) if rng is None else np.array(rng)
            points = to_range_1d(points, source=rng, target=[-1, 1]).flatten()
            rng = [-1, 1]
        else:
            if points is None:
                points = np.array(self.Geometry.master_coordinates())

        points, rng = np.array(points, dtype=float), np.array(rng, dtype=float)

        if nDIM > 1:
            points = atleastnd(points, 2, front=True)

        return points, rng

    def _parse_gauss_data(self, quad_dict: dict, key: Hashable) -> Iterable[Quadrature]:
        value: Union[Callable, str, dict] = quad_dict[key]

        if isinstance(value, Callable):
            qpos, qweight = value()
            quad = Quadrature(x=qpos, w=qweight)
            yield quad
        elif isinstance(value, str):
            for v in self._parse_gauss_data(quad_dict, value):
                yield v
        else:
            qpos, qweight = value
            quad = Quadrature(x=qpos, w=qweight)
            yield quad

    @property
    def frames(self) -> ndarray:
        """
        Returns local coordinate frames of the cells as a 3d NumPy float array,
        where the first axis runs along the cells of the block.
        """
        if not self.db.has_frames:
            if (nD := self.Geometry.number_of_spatial_dimensions) == 1:
                coords = self.source_coords()
                topo = self.topology().to_numpy()
                self.db.frames = frames_of_lines(coords, topo)
            elif nD == 2:
                coords = self.source_coords()
                topo = self.topology().to_numpy()
                self.db.frames = frames_of_surfaces(coords, topo)
            elif nD == 3:
                self.db.frames = self.source_frame()
            else:  # pragma: no cover
                raise TypeError(
                    "Invalid Geometry class. The 'number of spatial dimensions'"
                    " must be 1, 2 or 3."
                )
        return self.db.frames

    @frames.setter
    def frames(self, value: Union[FrameLike, ndarray]) -> None:
        self.db.frames = value

    def to_parquet(
        self, path: str, *args, fields: Iterable[str] = None, **kwargs
    ) -> None:
        """
        Saves the data of the database to a parquet file.

        Parameters
        ----------
        *args: tuple, Optional
            Positional arguments to specify fields.
        path: str
            Path of the file being created.
        fields: Iterable[str], Optional
            Valid field names to include in the parquet files.
        **kwargs: dict, Optional
            Keyword arguments forwarded to :func:`awkward.to_parquet`.
        """
        self.db.to_parquet(path, *args, fields=fields, **kwargs)

    @classmethod
    def from_parquet(cls, path: str) -> "PolyCell":
        """
        Saves the data of the database to a parquet file.

        Parameters
        ----------
        path: str
            Path of the file being created.
        """
        db_class = cls.data_class
        return cls(db=db_class.from_parquet(path))

    def to_triangles(self) -> ndarray:
        """
        Returns the topology as a collection of T3 triangles, represented
        as a 2d NumPy integer array, where the first axis runs along the
        triangles, and the second along the nodes.

        Only for 2d cells.
        """
        if self.Geometry.number_of_spatial_dimensions == 2:
            t = self.topology().to_numpy()
            return transform_topology(t, self.Geometry.trimap())
        else:
            raise NotImplementedError("This is only for 2d cells")

    def to_tetrahedra(self, flatten: Optional[bool] = True) -> ndarray:
        """
        Returns the topology as a collection of TET4 tetrahedra, represented
        as a 2d NumPy integer array, where the first axis runs along the
        tetrahedra, and the second along the nodes.

        Only for 3d cells.

        Parameters
        ----------
        flatten: bool, Optional
            If True, the topology is returned as a 2d array. If False, the
            length of the first axis equals the number of cells in the block,
            the length of the second axis equals the number of tetrahedra per
            cell.
        """
        if self.Geometry.number_of_spatial_dimensions == 3:
            t = self.topology().to_numpy()
            tetmap = self.Geometry.tetmap()
            tetra = transform_topology(t, tetmap)
            if flatten:
                return tetra
            else:
                nE = len(t)
                nT = len(tetmap)
                return tetra.reshape(nE, nT, 4)
        else:
            raise NotImplementedError("This is only for 3d cells")

    def to_simplices(self) -> Tuple[ndarray]:
        """
        Returns the cells of the block, refactorized into simplices. For cells
        of dimension 2, the returned 2d NumPy integer array represents 3-noded
        triangles, for 3d cells it is a collection of 4-noded tetrahedra.
        """
        NDIM: int = self.Geometry.number_of_spatial_dimensions
        if NDIM == 1:
            raise NotImplementedError
        elif NDIM == 2:
            return self.to_triangles()
        elif NDIM == 3:
            return self.to_tetrahedra()
        else:
            raise NotImplementedError

    def jacobian_matrix(
        self,
        *,
        pcoords: Optional[Union[Iterable[float], None]] = None,
        dshp: Optional[Union[ndarray, None]] = None,
        **kwargs,
    ) -> ndarray:
        """
        Returns the jacobian matrices of the cells in the block. The evaluation
        of the matrix is governed by the inputs in the following way:
        - if `dshp` is provided, it must be a matrix of shape function derivatives
          evaluated at the desired locations
        - the desired locations are specified through `pcoords`

        Parameters
        ----------
        pcoords: Iterable[float], Optional
            Locations of the evaluation points.
        dshp: numpy.ndarray, Optional
            3d array of shape function derivatives for the master cell,
            evaluated at some points. The array must have a shape of
            (nG, nNE, nD), where nG, nNE and nD are the number of evaluation
            points, nodes per cell and spatial dimensions.

        Returns
        -------
        numpy.ndarray
            A 4d array of shape (nE, nP, nD, nD), where nE, nP and nD
            are the number of elements, evaluation points and spatial
            dimensions. The number of evaluation points in the output
            is governed by the parameter 'dshp' or 'pcoords'.

        Note
        ----
        For 1d cells, the returned array is also 4 dimensional, with the last two
        axes being dummy.
        """
        ecoords = kwargs.get("_ec", self.local_coordinates())

        if dshp is None:
            x = (
                np.array(pcoords)
                if pcoords is not None
                else self.Geometry.master_coordinates()
            )
            dshp = self.Geometry.shape_function_derivatives(x)

        if self.Geometry.number_of_spatial_dimensions == 1:
            return jacobian_matrix_bulk_1d(dshp, ecoords)
        else:
            return jacobian_matrix_bulk(dshp, ecoords)

    def jacobian(
        self, *, jac: Optional[Union[ndarray, None]] = None, **kwargs
    ) -> Union[float, ndarray]:
        """
        Returns the jacobian determinant for one or more cells.

        Parameters
        ----------
        jac: numpy.ndarray, Optional
            One or more Jacobian matrices. Default is None.
        **kwargs: dict
            Forwarded to :func:`jacobian_matrix` if the jacobian
            is not provided by the parameter 'jac'.

        Returns
        -------
        float or numpy.ndarray
            Value of the Jacobian for one or more cells.

        See Also
        --------
        :func:`jacobian_matrix`
        """
        if jac is None:
            jac = self.jacobian_matrix(**kwargs)

        if self.Geometry.number_of_spatial_dimensions == 1:
            return jacobian_det_bulk_1d(jac)
        else:
            return np.linalg.det(jac)

    def flip(self) -> "PolyCell":
        """
        Reverse the order of nodes of the topology.
        """
        topo = self.topology().to_numpy()
        self.db.nodes = np.flip(topo, axis=1)
        return self

    def measures(self, *args, **kwargs) -> ndarray:
        """
        Returns generalized measures for each cell in the block.
        The generalized measure is length for 1d cells,
        area for 2d cells and volume for 3d cells.
        """
        NDIM: int = self.Geometry.number_of_spatial_dimensions
        if NDIM == 1:
            return self.lengths()
        elif NDIM == 2:
            return self.areas()
        elif NDIM == 3:
            return self.volumes()
        else:
            raise NotImplementedError

    def measure(self, *args, **kwargs) -> float:
        """
        Returns the net generalized measure for the cells in the
        block. The generalized measure is length for 1d cells,
        area for 2d cells and volume for 3d cells.
        """
        return np.sum(self.measures(*args, **kwargs))

    def thickness(self) -> ndarray:
        """
        Returns the thicknesses of the cells. If not set, a thickness
        of 1.0 is returned for each cell. Only for 2d cells.
        """
        if self.Geometry.number_of_spatial_dimensions == 2:
            if self.db.has_thickness:
                t = self.db.t
            else:
                t = np.ones(len(self), dtype=float)
            return t
        else:
            raise NotImplementedError("This is only for 2d cells")

    def length(self) -> float:
        """
        Returns the total length of the cells in the block.
        """
        if self.Geometry.number_of_spatial_dimensions == 1:
            return np.sum(self.lengths())
        else:
            raise NotImplementedError("This is only for 1d cells")

    def lengths(self) -> ndarray:
        """
        Returns the lengths as a NumPy array.
        """
        if self.Geometry.number_of_spatial_dimensions == 1:
            coords = self.container.source().coords()
            topo = self.topology().to_numpy()
            return lengths_of_lines(coords, topo)
        else:
            raise NotImplementedError("This is only for 1d cells")

    def area(self, *args, **kwargs) -> float:
        """
        Returns the total area of the cells in the database.
        Only for 2d entities.
        """
        if self.Geometry.number_of_spatial_dimensions == 2:
            return np.sum(self.areas(*args, **kwargs))
        else:
            raise NotImplementedError("This is only for 2d cells")

    def areas(self, *args, **kwargs) -> ndarray:
        """
        Returns the areas of the individuall cells in the database as
        a 1d float NumPy array.

        Note
        ----
        For 1d cells, the cross sectional areas are returned.
        """
        NDIM: int = self.Geometry.number_of_spatial_dimensions
        if NDIM == 1:
            if self.db.has_areas:
                return self.db.A
            else:
                return np.ones((len(self)))
        elif NDIM == 2:
            coords = self.source_coords()
            topo = self.topology().to_numpy()
            ecoords = cells_coords(coords[:, :2], topo)
            quad: Quadrature = next(
                self._parse_gauss_data(self.Geometry.quadrature, "geometry")
            )
            dshp = self.Geometry.shape_function_derivatives(quad.pos)
            return cell_measures(ecoords, dshp, quad.weight)
        else:
            raise NotImplementedError("This is only for 2d cells")

    def volume(self, *args, **kwargs) -> float:
        """
        Returns the volume of the cells in the database.
        """
        return np.sum(self.volumes(*args, **kwargs))

    def volumes(self, *args, **kwargs) -> ndarray:
        """
        Returns the volumes of the cells in the database.
        """
        NDIM: int = self.Geometry.number_of_spatial_dimensions
        if NDIM == 1:
            return self.lengths() * self.areas()
        elif NDIM == 2:
            areas = self.areas()
            t = self.thickness()
            return areas * t
        elif NDIM == 3:
            coords = self.source_coords()
            topo = self.topology().to_numpy()
            ecoords = cells_coords(coords, topo)
            quad: Quadrature = next(
                self._parse_gauss_data(self.Geometry.quadrature, "geometry")
            )
            dshp = self.Geometry.shape_function_derivatives(quad.pos)
            return cell_measures(ecoords, dshp, quad.weight)
        else:  # pragma: no cover
            raise NotImplementedError

    def source_points(self) -> PointCloud:
        """
        Returns the hosting pointcloud.
        """
        return self.container.source().points()

    def source_coords(self) -> ndarray:
        """
        Returns the coordinates of the hosting pointcloud.
        """
        if self.pointdata is not None:
            coords = self.pointdata.x
        else:
            coords = self.source().coords()
        return coords

    def source_frame(self) -> FrameLike:
        """
        Returns the frame of the hosting pointcloud.
        """
        return self.source().frame

    def points_of_cells(
        self,
        *,
        points: Optional[Union[float, Iterable, None]] = None,
        cells: Optional[Union[int, Iterable, None]] = None,
        rng: Optional[Union[Iterable, None]] = None,
    ) -> ndarray:
        """
        Returns the points of selected cells as a NumPy array. The returned
        array is three dimensional with a shape of (nE, nNE, nD), where `nE` is
        the number of cells in the block, `nNE` is the number of nodes per cell or
        the number of the points (if 'points' is specified) and nD stands for the
        number of spatial dimensions.

        Parameters
        ----------
        points: Optional[Union[float, Iterable, None]]
            Points defined in the domain of the master cell. If specified, global
            coordinates for each cell are calculated and returned for each cell.
            Default is `None`, in which case the locations of the nodes of the cells
            are used.
        cells: Optional[Union[int, Iterable, None]]
            BLock-local indices of the cells of interest, or `None` if all of the
            cells in the block are of interest. Default is `None`.
        rng: Optional[Union[Iterable, None]]
            For 1d cells only, it is possible to provide an iterable of length 2
            as an interval (or range) in which the argument `points` is to be understood.
            Default is `None`, in which case the `points` are expected in the range [-1, 1].
        """
        cells = self._get_cell_slicer(cells)

        NDIM: int = self.Geometry.number_of_spatial_dimensions
        coords = self.source_coords()
        topo = self.topology().to_numpy()[cells]
        ecoords = points_of_cells(coords, topo, centralize=False)

        if points is None:
            return ecoords

        points, rng = self._get_points_and_range(points, rng)

        if NDIM == 1:
            points = to_range_1d(points, source=rng, target=[0, 1]).flatten()
            res = pcoords_to_coords_1d(points, ecoords)  # (nE * nP, nD)
            nE = ecoords.shape[0]
            nP = points.shape[0]
            res = res.reshape(nE, nP, res.shape[-1])  # (nE, nP, nD)
            return res
        else:
            shp = self.Geometry.shape_function_values(points)
            if len(shp) == 3:  # variable metric cells
                shp = shp if len(shp) == 2 else shp[cells]
            return pcoords_to_coords(points, ecoords, shp)  # (nE, nP, nD)

    def local_coordinates(
        self, *, target: Optional[Union[str, CartesianFrame, None]] = None
    ) -> ndarray:
        """
        Returns local coordinates of the cells as a 3d float NumPy array.
        The returned array is three dimensional with a shape of (nE, nNE, 2),
        where `nE` is the number of cells in the block, `nNE` is the number of
        nodes per cell and 2 stands for the 2 spatial dimensions. The coordinates
        are centralized to the centers for each cell.

        Parameters
        ----------
        target: CartesianFrame, Optional
            A target frame. If provided, coordinates are returned in
            this frame, otherwise they are returned in the local frames
            of the cells. Default is None.
        """
        if isinstance(target, CartesianFrame):
            frames = target.show()
        else:
            frames = self.frames

        topo = self.topology().to_numpy()
        coords = self.source_coords()
        centers = self.loc_to_glob(self.Geometry.master_center())
        res = points_of_cells(
            coords, topo, local_axes=frames, centralize=True, centers=centers
        )

        if self.Geometry.number_of_spatial_dimensions == 2:
            return ascont(res[:, :, :2])
        else:
            return res

    def coords(self, *args, **kwargs) -> ndarray:
        """
        Alias for :func:`points_of_cells`, all arguments are forwarded.
        """
        return self.points_of_cells(*args, **kwargs)

    def topology(self) -> TopologyArray | None:
        """
        Returns the numerical representation of the topology of
        the cells as either a :class:`~sigmaepsilon.mesh.topoarray.TopologyArray`
        or `None` if the topology is not specified yet.
        """
        if self.db.has_nodes:
            return TopologyArray(self.db.nodes)
        else:  # pragma: no cover
            return None

    def rewire(
        self,
        imap: Optional[Union[MapLike, None]] = None,
        invert: Optional[bool] = False,
    ) -> "PolyCell":
        """
        Rewires the topology of the block according to the mapping
        described by the argument `imap`. The mapping of the j-th node
        of the i-th cell happens the following way:

        topology_new[i, j] = imap[topology_old[i, j]]

        The object is returned for continuation.

        Parameters
        ----------
        imap: MapLike, Optional
            Mapping from old to new node indices (global to local).
        invert: bool, Optional
            If `True` the argument `imap` describes a local to global
            mapping and an inversion takes place. In this case,
            `imap` must be a `numpy` array. Default is False.
        """
        if imap is None:
            imap = self.source().pointdata.id
        topo = self.topology().to_array().astype(int)
        topo = rewire(topo, imap, invert=invert).astype(int)
        self.db.nodes = topo
        return self

    def glob_to_loc(self, x: Union[Iterable, ndarray]) -> ndarray:
        """
        Returns the local coordinates of the input points for each
        cell in the block. The input 'x' can describe a single (1d array),
        or several positions at once (2d array).

        Notes
        -----
        This function is useful when detecting if two bodies touch each other or not,
        and if they do, where.

        Parameters
        ----------
        x: Iterable or numpy.ndarray
            A single point in 3d space as an 1d array, or a collection of points
            as a 2d array.

        Returns
        -------
        numpy.ndarray
            A NumPy array of shape (nE, nP, nD), where nP is the number of points in 'x',
            nE is the number of cells in the block and nD is the number of spatial dimensions.
        """
        raise NotImplementedError

    def loc_to_glob(
        self, x: Union[Iterable, ndarray], ec: Optional[Union[ndarray, None]] = None
    ) -> ndarray:
        """
        Returns the global coordinates of the input points for each
        cell in the block. The input 'x' can describe a single (1d array),
        or several local positions at once (2d array).

        Notes
        -----
        This function is useful when detecting if two bodies touch each other or not,
        and if they do, where.

        Parameters
        ----------
        x: Iterable or numpy.ndarray
            A single point as an 1d array, or a collection of points
            as a 2d array.
        ec: numpy.ndarray, Optional
            Element coordinates as a 3d array of shape (nE, nNE, nD).
            Default is None, in which case the global coordinates of the
            cells are used.

        Returns
        -------
        numpy.ndarray
            A NumPy array of shape (nE, nP, nD), where nP is the number of points in 'x',
            nE is the number of cells in the block and nD is the number of spatial dimensions.
        """
        nD = self.Geometry.number_of_spatial_dimensions
        if nD > 1:
            x = atleast2d(x, front=True)
        shp = self.Geometry.shape_function_values(x)  # (nP, nNE)
        if ec is None:
            ec = self.points_of_cells()
        return _loc_to_glob_bulk_(shp, ec)

    def pip(
        self,
        x: Union[Iterable, ndarray],
        tol: float = 1e-12,
        lazy: bool = True,
        k: int = 4,
    ) -> Union[bool, ndarray]:
        """
        Returns an 1d boolean integer array that tells if the points specified by 'x'
        are included in any of the cells in the block.

        Parameters
        ----------
        x: Iterable or numpy.ndarray
            The coordinates of the points that we want to investigate.
        tol: float, Optional
            Floating point tolerance for detecting boundaries. Default is 1e-12.
        lazy: bool, Optional
            If False, the ckeck is performed for all cells in the block. If True,
            it is used in combination with parameter 'k' and the check is only performed
            for the k nearest neighbours of the input points. Default is True.
        k: int, Optional
            The number of neighbours for the case when 'lazy' is true. Default is 4.

        Returns
        -------
        bool or numpy.ndarray
            A single or NumPy array of booleans for every input point.
        """
        NDIM: int = self.Geometry.number_of_spatial_dimensions
        if NDIM == 2:
            x = atleast2d(x, front=True)
            coords = self.source_coords()
            topo = self.to_triangles()
            ecoords = points_of_cells(coords, topo, centralize=False)
            if lazy:
                raise NotImplementedError
            else:
                pips = _pip_tri_bulk_(x, ecoords, tol)
            return np.squeeze(np.any(pips, axis=1))
        elif NDIM == 3:
            x = atleast2d(x, front=True)
            coords = self.source_coords()
            topo = self.to_tetrahedra(flatten=True)
            ecoords = points_of_cells(coords, topo, centralize=False)
            if lazy:
                centers = cell_centers_bulk(coords, topo)
                k = min(k, len(centers))
                knn = k_nearest_neighbours(centers, x, k=k)
                pips = _pip_tet_bulk_knn_(x, ecoords, knn, tol)
            else:
                pips = _pip_tet_bulk_(x, ecoords, tol)
            return np.squeeze(np.any(pips, axis=1))
        else:
            raise NotImplementedError

    def locate(
        self,
        x: Union[Iterable, ndarray],
        lazy: bool = True,
        tol: float = 1e-12,
        k: int = 4,
    ) -> Tuple[ndarray]:
        """
        Locates a set of points inside the cells of the block.

        Parameters
        ----------
        x: Iterable or numpy.ndarray
            The coordinates of the points that we want to investigate.
        tol: float, Optional
            Floating point tolerance for detecting boundaries. Default is 1e-12.
        lazy: bool, Optional
            If False, the ckeck is performed for all cells in the block. If True,
            it is used in combination with parameter 'k' and the check is only performed
            for the k nearest neighbours of the input points. Default is True.
        k: int, Optional
            The number of neighbours for the case when 'lazy' is true. Default is 4.

        Returns
        -------
        numpy.ndarray
            The indices of 'x' that are inside a cell of the block.
        numpy.ndarray
            The block-local indices of the cells that include the points with
            the returned indices.
        numpy.ndarray
            The parametric coordinates of the located points inside the including cells.
        """
        NDIM: int = self.Geometry.number_of_spatial_dimensions
        if NDIM == 3:
            x = atleast2d(x, front=True)

            coords = self.source_coords()
            topo = self.topology()
            topo_tet = self.to_tetrahedra(flatten=True)
            ecoords_tet = points_of_cells(coords, topo_tet, centralize=False)
            tetmap = self.Geometry.tetmap()

            # perform point-in-polygon test for tetrahedra
            if lazy:
                centers_tet = cell_centers_bulk(coords, topo_tet)
                k_tet = min(k, len(centers_tet))
                neighbours_tet = k_nearest_neighbours(centers_tet, x, k=k_tet)
                nat_tet = _glob_to_nat_tet_bulk_knn_(
                    x, ecoords_tet, neighbours_tet
                )  # (nP, kTET, 4)
                pips_tet = _pip_tet_bulk_nat_(nat_tet, tol)  # (nP, kTET)
            else:
                nat_tet = _glob_to_nat_tet_bulk_(x, ecoords_tet)  # (nP, nTET, 4)
                pips_tet = _pip_tet_bulk_nat_(nat_tet, tol)  # (nP, nTET)

            # locate the points that are inside any of the cells
            pip = np.squeeze(np.any(pips_tet, axis=1))  # (nP)
            i_source = np.where(pip)[0]  # (nP_)
            if lazy:
                points_to_tets, points_to_neighbours = _find_first_hits_knn_(
                    pips_tet[i_source], neighbours_tet[i_source]
                )
            else:
                points_to_tets, points_to_neighbours = _find_first_hits_(
                    pips_tet[i_source]
                )
            tets_to_cells = np.floor(np.arange(len(topo_tet)) / len(tetmap)).astype(int)
            i_target = tets_to_cells[points_to_tets]  # (nP_)

            # locate the cells that contain the points
            cell_tet_indices = np.tile(np.arange(tetmap.shape[0]), len(topo))[
                points_to_tets
            ]
            nat_tet = nat_tet[i_source]  # (nP_, nTET, 4)
            locations_target = _ntet_to_loc_bulk_(
                self.Geometry.master_coordinates(),
                nat_tet,
                tetmap,
                cell_tet_indices,
                points_to_neighbours,
            )

            return i_source, i_target, locations_target
        else:
            raise NotImplementedError

    def centers(self, target: Optional[Union[CartesianFrame, None]] = None) -> ndarray:
        """
        Returns the centers of the cells of the block as a 1d float
        NumPy array.

        Parameters
        ----------
        target: CartesianFrame, Optional
            A target frame. If provided, coordinates are returned in
            this frame, otherwise they are returned in the global frame.
            Default is None.
        """
        coords = self.source_coords()
        t = self.topology().to_numpy()
        centers = cell_centers_bulk(coords, t)
        if target:
            pc = PointCloud(centers, frame=self.source_frame())
            centers = pc.show(target)
        return centers

    def normals(self) -> ndarray:
        """
        Returns the normals of the cells as a 2d NumPy array.
        """
        if self.Geometry.number_of_spatial_dimensions == 2:
            return ascont(self.frames[:, 2, :])
        else:
            raise NotImplementedError("This is only available for 2d cells.")

    def unique_indices(self) -> ndarray:
        """
        Returns the indices of the points involved in the cells of the block
        as a 1d integer NumPy array.
        """
        return np.unique(self.topology())

    def points_involved(self) -> PointCloud:
        """
        Returns the points involved in the cells of the block.
        """
        return self.source_points()[self.unique_indices()]

    def detach_points_cells(self) -> Tuple[ndarray]:
        """
        Returns the detached coordinate and topology array of the block.
        """
        coords = self.container.source().coords()
        topo = self.topology().to_numpy()
        return detach_mesh_bulk(coords, topo)

    def to_vtk(self, detach: bool = False) -> Any:
        """
        Returns the block as a VTK object.

        Parameters
        ----------
        detach: bool, Optional
            Wether to detach the mesh or not. Default is False.
        """
        coords = self.container.source().coords()
        topo = self.topology().to_numpy()
        vtkid: int = self.Geometry.vtk_cell_id
        if detach:
            ugrid = mesh_to_vtk(*detach_mesh_bulk(coords, topo), vtkid)
        else:
            ugrid = mesh_to_vtk(coords, topo, vtkid)
        return ugrid

    if __haspyvista__:

        def to_pv(
            self, detach: bool = False
        ) -> Union[pv.UnstructuredGrid, pv.PolyData]:
            """
            Returns the block as a PyVista object.

            Parameters
            ----------
            detach: bool, Optional
                Wether to detach the mesh or not. Default is False.
            """
            return pv.wrap(self.to_vtk(detach=detach))

    def extract_surface(self, detach: bool = False) -> Tuple[ndarray]:
        """
        Extracts the surface of the object as a 2-tuple of NumPy arrays
        representing the coordinates and the topology of a triangulation.

        Parameters
        ----------
        detach: bool, Optional
            Wether to detach the mesh or not. Default is False.
        """

        if self.Geometry.number_of_spatial_dimensions == 3:
            coords = self.source_coords()
            pvs = self.to_pv(detach=False).extract_surface()
            s = pvs.triangulate().cast_to_unstructured_grid()
            topo = s.cells_dict[5]
            if detach:
                return s.points, topo
            else:
                coords = self.source_coords()
                imap = index_of_closest_point(coords, np.array(s.points, dtype=float))
                topo = rewire(topo, imap)
                return coords, topo
        else:
            raise NotImplementedError

    def boundary(self, detach: bool = False) -> Tuple[ndarray]:
        """
        Alias for :func:`extract_surface`.
        """
        if self.Geometry.number_of_spatial_dimensions == 3:
            return self.extract_surface(detach=detach)
        else:
            raise NotImplementedError

    def _rotate_(self, *args, **kwargs):
        # this is triggered upon transformations performed on the hosting pointcloud
        if self.has_frames:
            source_frame = self.container.source().frame
            new_frames = (
                CartesianFrame(self.frames, assume_cartesian=True)
                .rotate(*args, **kwargs)
                .show(source_frame)
            )
            self.frames = new_frames

    def __len__(self) -> int:
        return len(self.db)

    def __hasattr__(self, attr: str) -> Any:
        return attr in self.__dict__ or hasattr(self.db, attr)

    def __getattr__(self, attr: str) -> Any:
        if attr in self.__dict__:
            return self.__dict__[attr]
        try:
            return getattr(self.db, attr)
        except Exception:
            raise AttributeError(
                "'{}' object has no attribute \
                called {}".format(
                    self.__class__.__name__, attr
                )
            )

    def __getitem__(self, index: str) -> Any:
        try:
            return super().__getitem__(index)
        except Exception:
            try:
                return self.db.__getitem__(index)
            except Exception:
                raise TypeError(
                    "'{}' object is not "
                    "subscriptable".format(self.__class__.__name__)
                )

    def __setitem__(self, index: str, value: Any) -> None:
        if not isinstance(index, str):
            raise TypeError(f"Expected a string, got {type(index)}")

        if not (issequence(value) and len(value) == len(self.db)):
            raise ValueError(
                "The length of the provided data must match the number of cells in the block"
            )

        self.db[index] = value

    def __deepcopy__(self, memo: dict) -> "PolyCell":
        return self.__copy__(memo)

    def __copy__(self, memo: Optional[Union[dict, None]] = None) -> "PolyCell":
        cls = type(self)
        is_deep = memo is not None

        if is_deep:
            copy_function = lambda x: deepcopy(x, memo)
        else:
            copy_function = lambda x: x

        db = copy_function(self.db)

        pd = self.pointdata
        pd_copy = None
        if pd is not None:
            if is_deep:
                pd_copy = memo.get(id(pd), None)
            if pd_copy is None:
                pd_copy = copy_function(pd)

        result = cls(db=db, pointdata=pd_copy)
        if is_deep:
            memo[id(self)] = result

        result_dict = result.__dict__
        for k, v in self.__dict__.items():
            if not k in result_dict:
                setattr(result, k, copy_function(v))

        return result



# -*- coding: utf-8 -*-
from typing import Tuple

import numpy as np
from numpy import ndarray
from numba import njit

__cache = True


@njit(nogil=True, cache=__cache)
def edges_Q4(quads: ndarray) -> ndarray:
    nE = len(quads)
    edges = np.zeros((nE, 4, 2), dtype=quads.dtype)
    edges[:, 0, 0] = quads[:, 0]
    edges[:, 0, 1] = quads[:, 1]
    edges[:, 1, 0] = quads[:, 1]
    edges[:, 1, 1] = quads[:, 2]
    edges[:, 2, 0] = quads[:, 2]
    edges[:, 2, 1] = quads[:, 3]
    edges[:, 3, 0] = quads[:, 3]
    edges[:, 3, 1] = quads[:, 0]
    return edges


def edgeIds_T3() -> ndarray:
    return np.array([[0, 1], [1, 2], [2, 0]], dtype=int)


def edgeIds_TET4() -> ndarray:
    return np.array([[0, 1], [1, 2], [2, 0], [0, 3], [1, 3], [2, 3]], dtype=int)


@njit(nogil=True, cache=__cache)
def edges_TET4(topo: ndarray) -> ndarray:
    nE = len(topo)
    edges = np.zeros((nE, 6, 2), dtype=topo.dtype)
    edges[:, 0, 0] = topo[:, 0]
    edges[:, 0, 1] = topo[:, 1]
    edges[:, 1, 0] = topo[:, 1]
    edges[:, 1, 1] = topo[:, 2]
    edges[:, 2, 0] = topo[:, 2]
    edges[:, 2, 1] = topo[:, 0]
    edges[:, 3, 0] = topo[:, 0]
    edges[:, 3, 1] = topo[:, 3]
    edges[:, 4, 0] = topo[:, 1]
    edges[:, 4, 1] = topo[:, 3]
    edges[:, 5, 0] = topo[:, 2]
    edges[:, 5, 1] = topo[:, 3]
    return edges


@njit(nogil=True, cache=__cache)
def faces_TET4(topo: ndarray) -> ndarray:
    nE = len(topo)
    faces = np.zeros((nE, 4, 3), dtype=topo.dtype)
    faces[:, 0, 0] = topo[:, 0]
    faces[:, 0, 1] = topo[:, 1]
    faces[:, 0, 2] = topo[:, 3]
    faces[:, 1, 0] = topo[:, 1]
    faces[:, 1, 1] = topo[:, 2]
    faces[:, 1, 2] = topo[:, 3]
    faces[:, 2, 0] = topo[:, 2]
    faces[:, 2, 1] = topo[:, 0]
    faces[:, 2, 2] = topo[:, 3]
    faces[:, 3, 0] = topo[:, 0]
    faces[:, 3, 1] = topo[:, 2]
    faces[:, 3, 2] = topo[:, 1]
    return faces


def edgeIds_H8() -> ndarray:
    return np.array(
        [
            [0, 1],
            [1, 2],
            [2, 3],
            [3, 0],
            [4, 5],
            [5, 6],
            [6, 7],
            [7, 4],
            [0, 4],
            [1, 5],
            [2, 6],
            [3, 7],
        ],
        dtype=int,
    )


@njit(nogil=True, cache=__cache)
def edges_H8(topo: ndarray) -> ndarray:
    nE = len(topo)
    edges = np.zeros((nE, 12, 2), dtype=topo.dtype)
    edges[:, 0, 0] = topo[:, 0]
    edges[:, 0, 1] = topo[:, 1]
    edges[:, 1, 0] = topo[:, 1]
    edges[:, 1, 1] = topo[:, 2]
    edges[:, 2, 0] = topo[:, 2]
    edges[:, 2, 1] = topo[:, 3]
    edges[:, 3, 0] = topo[:, 3]
    edges[:, 3, 1] = topo[:, 0]
    edges[:, 4, 0] = topo[:, 4]
    edges[:, 4, 1] = topo[:, 5]
    edges[:, 5, 0] = topo[:, 5]
    edges[:, 5, 1] = topo[:, 6]
    edges[:, 6, 0] = topo[:, 6]
    edges[:, 6, 1] = topo[:, 7]
    edges[:, 7, 0] = topo[:, 7]
    edges[:, 7, 1] = topo[:, 4]
    edges[:, 8, 0] = topo[:, 0]
    edges[:, 8, 1] = topo[:, 4]
    edges[:, 9, 0] = topo[:, 1]
    edges[:, 9, 1] = topo[:, 5]
    edges[:, 10, 0] = topo[:, 2]
    edges[:, 10, 1] = topo[:, 6]
    edges[:, 11, 0] = topo[:, 3]
    edges[:, 11, 1] = topo[:, 7]
    return edges


@njit(nogil=True, cache=__cache)
def faces_H8(topo: ndarray) -> ndarray:
    nE = len(topo)
    faces = np.zeros((nE, 6, 4), dtype=topo.dtype)
    faces[:, 0, 0] = topo[:, 0]
    faces[:, 0, 1] = topo[:, 4]
    faces[:, 0, 2] = topo[:, 7]
    faces[:, 0, 3] = topo[:, 3]
    faces[:, 1, 0] = topo[:, 1]
    faces[:, 1, 1] = topo[:, 2]
    faces[:, 1, 2] = topo[:, 6]
    faces[:, 1, 3] = topo[:, 5]
    faces[:, 2, 0] = topo[:, 0]
    faces[:, 2, 1] = topo[:, 1]
    faces[:, 2, 2] = topo[:, 5]
    faces[:, 2, 3] = topo[:, 4]
    faces[:, 3, 0] = topo[:, 2]
    faces[:, 3, 1] = topo[:, 3]
    faces[:, 3, 2] = topo[:, 7]
    faces[:, 3, 3] = topo[:, 6]
    faces[:, 4, 0] = topo[:, 0]
    faces[:, 4, 1] = topo[:, 3]
    faces[:, 4, 2] = topo[:, 2]
    faces[:, 4, 3] = topo[:, 1]
    faces[:, 5, 0] = topo[:, 4]
    faces[:, 5, 1] = topo[:, 5]
    faces[:, 5, 2] = topo[:, 6]
    faces[:, 5, 3] = topo[:, 7]
    return faces


@njit(nogil=True, cache=__cache)
def edges_W6(topo: ndarray) -> ndarray:
    nE = len(topo)
    edges = np.zeros((nE, 9, 2), dtype=topo.dtype)
    edges[:, 0, 0] = topo[:, 0]
    edges[:, 0, 1] = topo[:, 1]
    edges[:, 1, 0] = topo[:, 1]
    edges[:, 1, 1] = topo[:, 2]
    edges[:, 2, 0] = topo[:, 2]
    edges[:, 2, 1] = topo[:, 0]
    edges[:, 3, 0] = topo[:, 3]
    edges[:, 3, 1] = topo[:, 4]
    edges[:, 4, 0] = topo[:, 4]
    edges[:, 4, 1] = topo[:, 5]
    edges[:, 5, 0] = topo[:, 5]
    edges[:, 5, 1] = topo[:, 3]
    edges[:, 6, 0] = topo[:, 0]
    edges[:, 6, 1] = topo[:, 3]
    edges[:, 7, 0] = topo[:, 1]
    edges[:, 7, 1] = topo[:, 4]
    edges[:, 8, 0] = topo[:, 2]
    edges[:, 8, 1] = topo[:, 5]
    return edges


@njit(nogil=True, cache=__cache)
def faces_W6(topo: ndarray) -> Tuple[ndarray, ndarray]:
    nE = len(topo)

    faces2 = np.zeros((nE, 2, 3), dtype=topo.dtype)
    faces2[:, 0, 0] = topo[:, 0]
    faces2[:, 0, 1] = topo[:, 2]
    faces2[:, 0, 2] = topo[:, 1]
    faces2[:, 1, 2] = topo[:, 3]
    faces2[:, 1, 0] = topo[:, 4]
    faces2[:, 1, 1] = topo[:, 5]

    faces4 = np.zeros((nE, 3, 4), dtype=topo.dtype)
    faces4[:, 0, 2] = topo[:, 0]
    faces4[:, 0, 3] = topo[:, 1]
    faces4[:, 0, 0] = topo[:, 4]
    faces4[:, 0, 1] = topo[:, 3]
    faces4[:, 1, 2] = topo[:, 1]
    faces4[:, 1, 3] = topo[:, 2]
    faces4[:, 1, 0] = topo[:, 5]
    faces4[:, 1, 1] = topo[:, 4]
    faces4[:, 2, 2] = topo[:, 2]
    faces4[:, 2, 3] = topo[:, 0]
    faces4[:, 2, 0] = topo[:, 3]
    faces4[:, 2, 1] = topo[:, 5]

    return faces4, faces2



# -*- coding: utf-8 -*-
from typing import Tuple, Optional, Union

import numpy as np
from numpy import ndarray
from numba import njit

from sigmaepsilon.math import atleast3d, repeat

from .utils.space import frames_of_lines
from .grid import rgridMT as grid


__cache = True


__all__ = ["mesh1d_uniform"]


@njit(nogil=True, parallel=False, fastmath=True, cache=__cache)
def _mesh1d_uniform_(
    coords: ndarray, topo: ndarray, eshape: ndarray, N: int, frames: ndarray
) -> Tuple[ndarray, ndarray, ndarray]:
    origo = np.zeros(1)
    subcoords_, subtopo_ = grid((1,), (N,), eshape, origo, 0)
    num_node_sub = len(subcoords_)
    N_new = len(coords) + len(topo) * (N - 1)
    coords_new = np.zeros((N_new, coords.shape[1]), dtype=coords.dtype)
    coords_new[: len(coords)] = coords
    frames_new = dict()
    topo_new = dict()
    cN = len(coords)  # node counter
    for i in range(topo.shape[0]):
        subtopo = subtopo_ + cN - 1
        subtopo[0, 0] = topo[i, 0]
        subtopo[-1, -1] = topo[i, -1]
        p1 = coords[topo[i, 0]]
        p2 = coords[topo[i, -1]]
        for j in range(1, num_node_sub - 1):
            p = p1 * (1 - subcoords_[j]) + p2 * subcoords_[j]
            coords_new[cN] = p
            cN += 1
        topo_new[i] = np.copy(subtopo)
        frames_new[i] = repeat(frames[i], N)
    return coords_new, topo_new, frames_new


def mesh1d_uniform(
    coords: ndarray,
    topo: ndarray,
    eshape: ndarray,
    *,
    N: Optional[int] = 2,
    refZ: Optional[Union[ndarray, None]] = None,
    return_frames: Optional[bool] = False,
    **__,
) -> Tuple[ndarray, ndarray]:
    """
    Returns the representation of a uniform 1d mesh as a tuple of numpy arrays.
    """
    frames = atleast3d(frames_of_lines(coords, topo, refZ))
    coords, topo, frames = _mesh1d_uniform_(coords, topo, eshape, N, frames)
    if return_frames:
        return coords, topo, frames
    return coords, topo



from numpy import ndarray

from sigmaepsilon.core.misc import get_index_suffix
from sigmaepsilon.core.exceptions import SigmaEpsilonException


class ArrayShapeMismatchError(SigmaEpsilonException):
    """
    An exception for the case for when one or more axes of an array does not
    have the desired length.
    """

    def __init__(
        self,
        message: str = None,
        *,
        arr_name: str = None,
        arr: ndarray = None,
        index: int = None,
        expected: int = None,
    ):
        if message is None:
            assert (
                arr is not None
                and arr_name is not None
                and index is not None
                and expected is not None
            )
            suffix = get_index_suffix(index)
            message = f"The {index}{suffix} axis of {arr_name} is {arr.shape[index]} but {expected} is expected."
        self.message = message
        super().__init__(message)

    @classmethod
    def check_axis_length(cls, arr: ndarray, arr_name: str, index: int, expected: int):
        if arr.shape[index] != expected:
            raise ArrayShapeMismatchError(
                arr=arr, arr_name=arr_name, index=index, expected=expected
            )



from .points import scatter_points_plotly
from .lines import plot_lines_plotly, scatter_lines_plotly
from .tri import triplot_plotly

__all__ = [
    "scatter_lines_plotly",
    "scatter_points_plotly",
    "plot_lines_plotly",
    "triplot_plotly",
]



from copy import copy, deepcopy
from typing import (
    Iterator,
    Union,
    Hashable,
    Collection,
    Iterable,
    Tuple,
    Any,
    Generic,
    Optional,
    Callable,
)
from collections import defaultdict
import functools
import warnings

from numpy import ndarray
import numpy as np
from scipy.sparse import spmatrix, csr_matrix as csr_scipy, diags as scipy_diags
import awkward as ak
from meshio import Mesh as MeshioMesh

from sigmaepsilon.deepdict import DeepDict
from sigmaepsilon.core.warning import SigmaEpsilonPerformanceWarning
from sigmaepsilon.math.linalg.sparse import csr_matrix, JaggedArray
from sigmaepsilon.math.linalg import Vector, ReferenceFrame as FrameLike
from sigmaepsilon.math import atleast1d, minmax

from ..typing import (
    PolyDataProtocol as PDP,
    PolyCellProtocol,
    PolyDataLike,
    PointDataLike,
    PolyCellLike,
)

from .akwrapper import AkWrapper
from .pointdata import PointData
from .polycell import PolyCell
from .polycell import PolyCell
from ..space import CartesianFrame, PointCloud
from ..indexmanager import IndexManager
from ..topoarray import TopologyArray

from ..utils.topology.topo import inds_to_invmap_as_dict, remap_topo_1d
from ..utils.utils import (
    cells_coords,
    cells_around,
    cell_centers_bulk,
    explode_mesh_data_bulk,
    nodal_distribution_factors,
)
from ..utils.knn import k_nearest_neighbours as KNN
from ..cells import T3 as Triangle
from ..utils.space import (
    index_of_closest_point,
    index_of_furthest_point,
    frames_of_surfaces,
)
from ..utils.topology import (
    nodal_adjacency,
    detach_mesh_data_bulk,
    detach_mesh_bulk,
    cells_at_nodes,
)
from ..helpers import importers, exporters, plotters
from ..config import __hasvtk__, __haspyvista__, __hask3d__, __hasmatplotlib__

if __hasvtk__:
    import vtk

if __hask3d__:
    import k3d

if __haspyvista__:
    import pyvista as pv

    pyVistaLike = Union[pv.PolyData, pv.PointGrid, pv.UnstructuredGrid]
else:  # pragma: no cover
    pyVistaLike = Any


VectorLike = Union[Vector, ndarray]

__all__ = ["PolyData"]


class PolyData(DeepDict, Generic[PointDataLike, PolyCellLike]):
    """
    A class to handle complex polygonal meshes.

    The `PolyData` class is the most important class in the library
    and a backbone of all mesh classes.

    The implementation is based on the `Awkward` library, which provides
    memory-efficient, numba-jittable data classes to deal with dense, sparse,
    complete or incomplete data. These data structures are managed in pure
    Python by the `DeepDict` class.

    The class accepts several kinds of inputs, allowing for a wide range of
    possible use cases. The fastes way to create a PolyData is from predefined
    `PointData` and `CellData` instances, defined separately.

    Parameters
    ----------
    pd: Union[PointData, CellData], Optional
        A PolyData or a CellData instance. Dafault is None.
    cd: CellData, Optional
        A CellData instance, if the first argument is provided. Dafault is None.

    Examples
    --------
    To create a simple cube:

    >>> from sigmaepsilon.mesh import PolyData, PointData
    >>> from sigmaepsilon.mesh.grid import grid
    >>> from sigmaepsilon.mesh.space import StandardFrame
    >>> from sigmaepsilon.mesh.cells import H27
    >>> size = Lx, Ly, Lz = 100, 100, 100
    >>> shape = nx, ny, nz = 10, 10, 10
    >>> coords, topo = grid(size=size, shape=shape, eshape='H27')
    >>> frame = StandardFrame(dim=3)
    >>> mesh = PolyData(pd=PointData(coords=coords, frame=frame))
    >>> mesh['A']['Part1'] = PolyData(cd=H27(topo=topo[:10], frames=frame))
    >>> mesh['A']['Part2'] = PolyData(cd=H27(topo=topo[10:-10], frames=frame))
    >>> mesh['A']['Part3'] = PolyData(cd=H27(topo=topo[-10:], frames=frame))
    >>> mesh.plot()  # doctest: +SKIP

    Load a mesh from a PyVista object:

    >>> from pyvista import examples
    >>> from sigmaepsilon.mesh import PolyData
    >>> bunny = examples.download_bunny_coarse()
    >>> mesh = PolyData.from_pv(bunny)

    Read from a .vtk file:

    >>> from sigmaepsilon.mesh import PolyData
    >>> from sigmaepsilon.mesh.downloads import download_stand
    >>> vtkpath = download_stand()
    >>> mesh = PolyData.read(vtkpath)

    See also
    --------
    :class:`~sigmaepsilon.mesh.data.trimesh.TriMesh`
    :class:`~sigmaepsilon.mesh.data.pointdata.PointData`
    :class:`~sigmaepsilon.mesh.data.celldata.CellData`
    """

    _point_array_class_ = PointCloud
    _point_class_ = PointData
    _frame_class_ = CartesianFrame
    _pv_config_key_ = ("pv", "default")
    _k3d_config_key_ = ("k3d", "default")

    def __init__(
        self,
        pd: Optional[Union[PointData, PolyCell, None]] = None,
        cd: Optional[Union[PolyCell, None]] = None,
        *args,
        **kwargs,
    ):
        self._reset_point_data()
        self._reset_cell_data()
        self._parent = None
        self._config = None
        self._cid2bid = None  # maps cell indices to block indices
        self._bid2b = None  # maps block indices to block addresses
        self._pointdata = None
        self._celldata = None
        self._init_config_()

        self.point_index_manager = IndexManager()
        self.cell_index_manager = IndexManager()

        if isinstance(pd, PointData):
            self.pointdata = pd
            if isinstance(cd, PolyCell):
                self.celldata = cd
        elif isinstance(pd, PolyCell):
            self.celldata = pd
            if isinstance(cd, PointData):
                self.pointdata = cd
        elif isinstance(cd, PolyCell):
            self.celldata = cd

        pidkey = self.__class__._point_class_._dbkey_id_

        if self.pointdata is not None:
            if self.pd.has_id:
                if self.celldata is not None:
                    imap = self.pd.id
                    self.cd.rewire(imap=imap, invert=True)
            N = len(self.pointdata)
            GIDs = self.root.pim.generate_np(N)
            self.pd[pidkey] = GIDs
            self.pd.container = self

        if self.celldata is not None:
            N = len(self.celldata.db)
            GIDs = self.root.cim.generate_np(N)
            self.cd.db.id = GIDs
            try:
                pd = self.source().pd
            except Exception:
                pd = None
            self.cd.pd = pd
            self.cd.container = self

        super().__init__(*args, **kwargs)

        if self.celldata is not None:
            self.celltype = self.celldata.__class__
            self.celldata.container = self

    def __deepcopy__(self, memo):
        return self.__copy__(memo)

    def __copy__(self, memo=None):
        cls = type(self)
        is_deep = memo is not None

        if is_deep:
            copy_function = lambda x: deepcopy(x, memo)
        else:
            copy_function = lambda x: x

        frame_cls = self._frame_class_

        # initialize result
        if self.frame is not None:
            f = self.frame
            ax = copy_function(f.axes)
            if is_deep:
                memo[id(f.axes)] = ax
            frame = frame_cls(ax)
        else:
            frame = None
        result = cls(frame=frame)
        if is_deep:
            memo[id(self)] = result

        # self
        if self.pointdata is not None:
            result.pointdata = copy_function(self.pointdata)

        if self.celldata is not None:
            result.celldata = copy_function(self.celldata)

        for k, v in self.items():
            if not isinstance(v, PolyData):
                result[k] = copy_function(v)

        result_dict = result.__dict__

        for k, v in self.__dict__.items():
            if not k in result_dict:
                setattr(result, k, copy_function(v))

        # children
        l0 = len(self.address)
        for b in self.blocks(inclusive=False, deep=True):
            pd, cd, bframe = None, None, None
            addr = b.address

            if len(addr) > l0:
                # pointdata
                if b.pointdata is not None:
                    pd = copy_function(b.pd)
                    # block frame
                    f = b.frame
                    ax = copy_function(f.axes)

                    if is_deep:
                        memo[id(f.axes)] = ax

                    bframe = frame_cls(ax)

                # celldata
                if b.celldata is not None:
                    cd = copy_function(b.cd)

                # mesh object
                pd_result = PolyData(pd, cd, frame=bframe)
                result[addr[l0:]] = pd_result

                # other data
                for k, v in b.items():
                    if not isinstance(v, PolyData):
                        pd_result[k] = copy_function(v)

                pd_result_dict = pd_result.__dict__

                for k, v in b.__dict__.items():
                    if not k in pd_result_dict:
                        setattr(pd_result, k, copy_function(v))

        return result

    def copy(self: PolyDataLike) -> PolyDataLike:
        """
        Returns a shallow copy.
        """
        return copy(self)

    def deepcopy(self: PolyDataLike) -> PolyDataLike:
        """
        Returns a deep copy.
        """
        return deepcopy(self)

    def __getitem__(self: PolyDataLike, key) -> PolyDataLike:
        return super().__getitem__(key)

    @property
    def pointdata(self) -> PointDataLike:
        """
        Returns the attached pointdata.
        """
        return self._pointdata

    @pointdata.setter
    def pointdata(self, pd: Optional[PointData]) -> None:
        """
        Returns the attached pointdata.
        """
        if pd is not None and not isinstance(pd, PointData):
            raise TypeError("Value must be a PointData instance.")
        self._pointdata = pd
        if isinstance(pd, PointData):
            self._pointdata.container = self

    @property
    def pd(self) -> PointDataLike:
        """
        Returns the attached pointdata.
        """
        return self.pointdata

    @property
    def celldata(self) -> PolyCellLike:
        """
        Returns the attached celldata.
        """
        return self._celldata

    @celldata.setter
    def celldata(self, cd: Optional[PolyCell]) -> None:
        """
        Returns the attached celldata.
        """
        if cd is not None and not isinstance(cd, PolyCell):
            raise TypeError("Value must be a PolyCell instance.")
        self._celldata = cd
        if isinstance(cd, PolyCell):
            self._celldata.container = self

    @property
    def cd(self) -> PolyCellLike:
        """
        Returns the attached celldata.
        """
        return self.celldata

    def lock(
        self: PolyDataLike, create_mappers: Optional[bool] = False
    ) -> PolyDataLike:
        """
        Locks the layout. If a `PolyData` instance is locked,
        missing keys are handled the same way as they would've been handled
        if it was a `dict`. Also, setting or deleting items in a locked
        dictionary and not possible and you will experience an error upon
        trying.

        The object is returned for continuation.

        Parameters
        ----------
        create_mappers: bool, Optional
            If True, some mappers are generated to speed up certain types of
            searches, like finding a block containing cells based on their
            indices.
        """
        if create_mappers and self._cid2bid is None:
            bid2b, cid2bid = self._create_mappers_()
            self._cid2bid = cid2bid  # maps cell indices to block indices
            self._bid2b = bid2b  # maps block indices to block addresses
        self._locked = True
        return self

    def unlock(self: PolyDataLike) -> PolyDataLike:
        """
        Releases the layout. If a `sigmaepsilon.mesh` instance is not locked,
        a missing key creates a new level in the layout, also setting and
        deleting items becomes an option. Additionally, mappers created with
        the call `generate_cell_mappers` are deleted.

        The object is returned for continuation.
        """
        self._locked = False
        self._cid2bid = None  # maps cell indices to block indices
        self._bid2b = None  # maps block indices to block addresses
        return self

    def blocks_of_cells(self, i: Optional[Union[int, Iterable, None]] = None) -> dict:
        """
        Returns a dictionary that maps cell indices to blocks.
        """
        assert self.is_root(), "This must be called on the root object."

        if self._cid2bid is None:
            warnings.warn(
                "Calling 'obj.lock(create_mappers=True)' creates additional"
                " mappers that make lookups like this much more efficient. "
                "See the doc of the sigmaepsilon.mesh library for more details.",
                SigmaEpsilonPerformanceWarning,
            )
            bid2b, cid2bid = self._create_mappers_()
        else:
            cid2bid = self._cid2bid
            bid2b = self._bid2b

        if i is None:
            return {cid: bid2b[bid] for cid, bid in cid2bid.items()}

        cids = atleast1d(i)
        bids = [cid2bid[cid] for cid in cids]
        cid2b = {cid: bid2b[bid] for cid, bid in zip(cids, bids)}

        return cid2b

    def _create_mappers_(self) -> Tuple[dict, dict]:
        """
        Generates mappers between cells and blocks to speed up some
        queries. This can only be called on the root object.
        The object is returned for continuation.
        """
        assert self.is_root(), "This must be called on the root object."
        bid2b = {}  # block index to block address
        cids = []  # cell indices
        bids = []  # block infices of cells
        for bid, b in enumerate(self.cellblocks(inclusive=True)):
            b.id = bid
            bid2b[bid] = b
            cids.append(b.cd.id)
            bids.append(np.full(len(b.cd), bid))
        cids = np.concatenate(cids)
        bids = np.concatenate(bids)
        cid2bid = {cid: bid for cid, bid in zip(cids, bids)}
        return bid2b, cid2bid

    @classmethod
    def read(cls: PolyDataLike, *args, **kwargs) -> PolyDataLike:
        """
        Reads from a file using PyVista.

        Example
        -------
        Download a .vtk file and read it:

        >>> from sigmaepsilon.mesh import PolyData
        >>> from sigmaepsilon.mesh.downloads import download_stand
        >>> vtkpath = download_stand(read=False)
        >>> mesh = PolyData.read(vtkpath)
        """
        return cls.from_pv(pv.read(*args, **kwargs))

    @classmethod
    def from_meshio(cls: PolyDataLike, mesh: MeshioMesh) -> PolyDataLike:
        """
        Returns a :class:`~sigmaepsilon.mesh.polydata.PolyData` instance from
        a :class:`meshio.Mesh` instance.

        .. note::
            See https://github.com/nschloe/meshio for formats supported by
            ``meshio``. Be sure to install ``meshio`` with ``pip install
            meshio`` if you wish to use it.
        """
        importer: Callable = importers["meshio"]
        return importer(mesh)

    @classmethod
    def from_pv(cls: PolyDataLike, pvobj) -> PolyDataLike:
        """
        Returns a :class:`~sigmaepsilon.mesh.polydata.PolyData` instance from
        a :class:`pyvista.PolyData` or a :class:`pyvista.UnstructuredGrid`
        instance.

        .. note::
            See https://github.com/pyvista/pyvista for more examples with
            ``pyvista``. Be sure to install ``pyvista`` with ``pip install
            pyvista`` if you wish to use it.

        Example
        -------
        >>> from pyvista import examples
        >>> from sigmaepsilon.mesh import PolyData
        >>> bunny = examples.download_bunny_coarse()
        >>> mesh = PolyData.from_pv(bunny)
        """
        importer: Callable = importers["PyVista"]
        return importer(pvobj)

    def to_dataframe(
        self,
        *args,
        point_fields: Optional[Union[Iterable[str], None]] = None,
        cell_fields: Optional[Union[Iterable[str], None]] = None,
        **kwargs,
    ) -> Any:
        """
        Returns the data contained within the mesh to pandas dataframes.

        Parameters
        ----------
        point_fields: Iterable[str], Optional
            A list of keys that might identify data in a database for the
            points in the mesh. Default is None.
        cell_fields: Iterable[str], Optional
            A list of keys that might identify data in a database for the
            cells in the mesh. Default is None.

        Example
        -------
        >>> from sigmaepsilon.mesh.downloads import download_stand
        >>> mesh = download_stand(read=True)
        >>> _ = mesh.to_dataframe(point_fields=mesh.pd.fields)
        """
        ak_pd, ak_cd = self.to_ak(
            *args, point_fields=point_fields, cell_fields=cell_fields
        )
        return ak.to_dataframe(ak_pd, **kwargs), ak.to_dataframe(ak_cd, **kwargs)

    def to_parquet(
        self,
        path_pd: str,
        path_cd: str,
        *args,
        point_fields: Optional[Union[Iterable[str], None]] = None,
        cell_fields: Optional[Union[Iterable[str], None]] = None,
        **kwargs,
    ) -> None:
        """
        Saves the data contained within the mesh to parquet files.

        Parameters
        ----------
        path_pd: str
            File path for point-related data.
        path_cd: str
            File path for cell-related data.
        point_fields: Iterable[str], Optional
            A list of keys that might identify data in a database for the
            points in the mesh. Default is None.
        cell_fields: Iterable[str], Optional
            A list of keys that might identify data in a database for the
            cells in the mesh. Default is None.

        Example
        -------
        >>> from sigmaepsilon.mesh.downloads import download_stand
        >>> mesh = download_stand(read=True)
        >>> _ = mesh.to_parquet('pd.parquet', 'cd.parquet', point_fields=mesh.pd.fields)
        """
        ak_pd, ak_cd = self.to_ak(
            *args, point_fields=point_fields, cell_fields=cell_fields
        )
        ak.to_parquet(ak_pd, path_pd, **kwargs)
        ak.to_parquet(ak_cd, path_cd, **kwargs)

    def to_ak(
        self,
        *args,
        point_fields: Optional[Union[Iterable[str], None]] = None,
        cell_fields: Optional[Union[Iterable[str], None]] = None,
        **__,
    ) -> Tuple[ak.Array]:
        """
        Returns the data contained within the mesh as a tuple of two
        Awkward arrays.

        Parameters
        ----------
        point_fields: Iterable[str], Optional
            A list of keys that might identify data in a database for the
            points in the mesh. Default is None.
        cell_fields: Iterable[str], Optional
            A list of keys that might identify data in a database for the
            cells in the mesh. Default is None.

        Example
        -------
        >>> from sigmaepsilon.mesh.downloads import download_stand
        >>> mesh = download_stand(read=True)
        >>> _ = mesh.to_ak(point_fields=mesh.pd.fields)
        """
        lp, lc = self.to_lists(
            *args, point_fields=point_fields, cell_fields=cell_fields
        )
        return ak.from_iter(lp), ak.from_iter(lc)

    def to_lists(
        self,
        *,
        point_fields: Optional[Union[Iterable[str], None]] = None,
        cell_fields: Optional[Union[Iterable[str], None]] = None,
    ) -> Tuple[list]:
        """
        Returns data of the object as a tuple of lists. The first is a list
        of point-related, the other one is cell-related data. Unless specified
        by 'fields', all data is returned from the pointcloud and the related
        cells of the mesh.

        Parameters
        ----------
        point_fields: Iterable[str], Optional
            A list of keys that might identify data in a database for the
            points in the mesh. Default is None.
        cell_fields: Iterable[str], Optional
            A list of keys that might identify data in a database for the
            cells in the mesh. Default is None.

        Example
        -------
        >>> from sigmaepsilon.mesh.downloads import download_stand
        >>> mesh = download_stand(read=True)
        >>> _ = mesh.to_lists(point_fields=mesh.pd.fields)
        """
        # handle points
        blocks = self.pointblocks(inclusive=True, deep=True)
        if point_fields is not None:

            def foo(b):
                pdb = b.pd.db
                db = {}
                for f in point_fields:
                    if f in pdb.fields:
                        db[f] = pdb[f]
                    else:
                        raise KeyError(f"Point field {f} not found.")
                w = AkWrapper(fields=db)
                return w.db.to_list()

        else:

            def foo(b):
                return b.pd.db.to_list()

        lp = list(map(foo, blocks))
        lp = functools.reduce(lambda a, b: a + b, lp)
        # handle cells
        blocks = self.cellblocks(inclusive=True, deep=True)
        if cell_fields is not None:

            def foo(b):
                cdb = b.cd.db
                db = {}
                for f in cell_fields:
                    if f in cdb.fields:
                        db[f] = cdb[f]
                    else:
                        raise KeyError(f"Cell field {f} not found.")
                cd = AkWrapper(fields=db)
                return cd.db.to_list()

        else:

            def foo(b):
                return b.cd.db.to_list()

        lc = list(map(foo, blocks))
        lc = functools.reduce(lambda a, b: a + b, lc)
        return lp, lc

    @property
    def config(self) -> DeepDict:
        """
        Returns the configuration object.

        Returns
        -------
        :class:`linkeddeepdict.LinkedDeepDict`
            The configuration object.

        Example
        -------
        >>> from sigmaepsilon.mesh.downloads import download_stand
        >>> mesh = download_stand(read=True)

        To set configuration values related to plotting with `pyVista`,
        do the following:

        >>> mesh.config['pyvista', 'plot', 'color'] = 'red'
        >>> mesh.config['pyvista', 'plot', 'style'] = 'wireframe'

        Then, when it comes to plotting, you can specify your configuration
        with the `config_key` keyword argument:

        >>> mesh.pvplot(config_key=('pyvista', 'plot'))  # doctest: +SKIP

        This way, you can store several different configurations for
        different scenarios.
        """
        return self._config

    def _init_config_(self):
        self._config = DeepDict()
        key = self.__class__._pv_config_key_
        self.config[key]["show_edges"] = True

    @property
    def pim(self) -> "IndexManager":
        return self.point_index_manager

    @property
    def cim(self) -> "IndexManager":
        return self.cell_index_manager

    @property
    def parent(self: PolyDataLike) -> PolyDataLike:
        """Returns the parent of the object."""
        return self._parent

    @parent.setter
    def parent(self, value: PolyDataLike) -> None:
        """Sets the parent."""
        self._parent = value

    def is_source(self, key: Optional[Union[str, None]] = None) -> bool:
        """
        Returns `True`, if the object is a valid source of data
        specified by `key`.

        Parameters
        ----------
        key: str
            A valid key to the PointData of the mesh. If not specified
            the key is the key used for storing coorindates.
        """
        key = PointData._dbkey_x_ if key is None else key
        return self.pointdata is not None and key in self.pointdata.fields

    def source(
        self, key: Optional[Union[str, None]] = None
    ) -> Union[PDP[PointDataLike, PolyCellLike], None]:
        """
        Returns the closest (going upwards in the hierarchy) block that holds
        on to data with a certain field name. If called without arguments,
        it is looking for a block with a valid pointcloud, definition, otherwise
        the field specified by the argument `key`.

        Parameters
        ----------
        key: str
            A valid key in any of the blocks with data. Default is None.
        """
        if self.is_source(key):
            return self
        else:
            if self.is_root():
                return None
            else:
                return self.parent.source(key=key)

    def blocks(
        self: PolyDataLike,
        *,
        inclusive: bool = False,
        blocktype: Any = None,
        deep: bool = True,
        **__,
    ) -> Collection[PolyDataLike]:
        """
        Returns an iterable over nested blocks.

        Parameters
        ----------
        inclusive: bool, Optional
            Whether to include the object the call was made upon.
            Default is False.
        blocktype: Any, Optional
            A required type. Default is None, which means theat all
            subclasses of the PolyData class are accepted. Default is None.
        deep: bool, Optional
            If True, parsing goes into deep levels. If False, only the level
            of the current object is handled.

        Yields
        ------
        Any
            A PolyData instance. The actual type depends on the 'blocktype'
            parameter.
        """
        dtype = PolyData if blocktype is None else blocktype
        return self.containers(inclusive=inclusive, dtype=dtype, deep=deep)

    def pointblocks(
        self, *args, **kwargs
    ) -> Iterable[PDP[PointDataLike, PolyCellLike]]:
        """
        Returns an iterable over blocks with PointData. All arguments
        are forwarded to :func:`blocks`.

        Yields
        ------
        Any
            A PolyData instance with a PointData.

        See also
        --------
        :func:`blocks`
        :class:`~sigmaepsilon.mesh.data.pointdata.PointData`
        """
        return filter(lambda i: i.pd is not None, self.blocks(*args, **kwargs))

    def cellblocks(self, *args, **kwargs) -> Iterable[PDP[PointDataLike, PolyCellLike]]:
        """
        Returns an iterable over blocks with CellData. All arguments
        are forwarded to :func:`blocks`.

        Yields
        ------
        Any
            A CellData instance with a CellData.

        See also
        --------
        :func:`blocks`
        :class:`~sigmaepsilon.mesh.data.celldata.CellData`
        """
        return filter(lambda i: i.cd is not None, self.blocks(*args, **kwargs))

    @property
    def point_fields(self) -> Iterable[str]:
        """
        Returns the fields of all the pointdata of the object.

        Returns
        -------
        numpy.ndarray
            NumPy array of data keys.
        """
        pointblocks = list(self.pointblocks())
        m = map(lambda pb: pb.pointdata.fields, pointblocks)
        return np.unique(np.array(list(m)).flatten())

    @property
    def cell_fields(self) -> Iterable[str]:
        """
        Returns the fields of all the celldata of the object.

        Returns
        -------
        numpy.ndarray
            NumPy array of data keys.
        """
        cellblocks = list(self.cellblocks())
        m = map(lambda cb: cb.celldata.fields, cellblocks)
        return np.unique(np.array(list(m)).flatten())

    @property
    def frame(self) -> FrameLike:
        """Returns the frame of the underlying pointcloud."""
        result = None

        if self.pd is not None:
            if self.pd.has_x:
                result = self.pd.frame

        if result is None:
            if self.parent is not None:
                result = self.parent.frame

        # If the frame is still None, it means that the entire mesh
        # has no frame, not even the root object. In this case we assign
        # a default frame to the root.
        if result is None:
            result = CartesianFrame()

        return result

    def _reset_point_data(self):
        self.pointdata = None
        self.cell_index_manager = None

    def _reset_cell_data(self):
        self.celldata = None
        self.celltype = None

    def rewire(
        self: PolyDataLike,
        deep: Optional[bool] = True,
        imap: Optional[Union[ndarray, None]] = None,
        invert: Optional[bool] = False,
    ) -> PolyDataLike:
        """
        Rewires topology according to the index mapping of the source object.

        Parameters
        ----------
        deep: bool, Optional
            If `True`, the action propagates down. Default is True.
        imap: numpy.ndarray, Optional
            Index mapper. Either provided as a numpy array, or it gets
            fetched from the database. Default is None.
        invert: bool, Optional
            A flag to indicate wether the provided index map should be
            inverted or not. Default is False.

        Notes
        -----
        Unless node numbering was modified, subsequent executions have
        no effect after once called.

        Returns
        -------
        :class:`~sigmaepsilon.mesh.polydata.PolyData`
            Returnes the object instance for continuitation.
        """
        if not deep:
            if self.cd is not None:
                if imap is not None:
                    self.cd.rewire(imap=imap, invert=invert)
                else:
                    imap = self.source().pointdata.id
                    self.cd.rewire(imap=imap, invert=False)
        else:
            if imap is not None:
                [
                    cb.rewire(imap=imap, deep=False, invert=invert)
                    for cb in self.cellblocks(inclusive=True)
                ]
            else:
                [
                    cb.rewire(deep=False, invert=invert)
                    for cb in self.cellblocks(inclusive=True)
                ]
        return self

    def to_standard_form(
        self: PolyDataLike,
        inplace: Optional[bool] = True,
        default_point_fields: Optional[Union[dict, None]] = None,
        default_cell_fields: Optional[Union[dict, None]] = None,
    ) -> PolyDataLike:
        """
        Transforms the problem to standard form, which means
        a centralized pointdata and regular cell indices.

        Notes
        -----
        Some operation might work better if the layout of the mesh
        admits the standard form.

        Parameters
        ----------
        inplace: bool, Optional
            Performs the operations inplace. Default is True.
        default_point_fields: dict, Optional
            A dictionary to define default values for missing fields
            for point related data. If not specified, the default
            is `numpy.nan`.
        default_cell_fields: dict, Optional
            A dictionary to define default values for missing fields
            for cell related data. If not specified, the default
            is `numpy.nan`.
        """
        assert self.is_root(), "This must be called on he root object!"

        if not inplace:
            return deepcopy(self).to_standard_form(inplace=True)

        # merge points and point related data
        # + decorate the points with globally unique ids
        dpf = defaultdict(lambda: np.nan)
        if isinstance(default_point_fields, dict):
            dpf.update(default_point_fields)

        pim = IndexManager()
        pointtype = self.__class__._point_class_
        pointblocks = list(self.pointblocks(inclusive=True, deep=True))
        m = map(lambda pb: pb.pointdata.fields, pointblocks)
        fields = set(np.concatenate(list(m)))
        frame = self.frame

        data = {f: [] for f in fields}
        for pb in pointblocks:
            id = pim.generate_np(len(pb.pointdata))
            pb.pointdata.id = id
            pb.pd.x = PointCloud(pb.pd.x, frame=pb.frame).show(frame)
            for f in fields:
                if f in pb.pd.fields:
                    data[f].append(pb.pointdata[f].to_numpy())
                else:
                    data[f].append(np.full(len(pb.pd), dpf[f]))

        X = np.concatenate(data.pop(PointData._dbkey_x_), axis=0)

        point_fields = {}
        for f in data.keys():
            point_fields[f] = np.concatenate(data[f], axis=0)

        self.pointdata = pointtype(
            coords=X, frame=frame, fields=point_fields, container=self
        )

        # merge cells and cell related data
        # + rewire the topology based on the ids set in the previous block
        dcf = defaultdict(lambda: np.nan)
        if isinstance(default_cell_fields, dict):
            dcf.update(default_cell_fields)

        cim = IndexManager()
        cellblocks = list(self.cellblocks(inclusive=True, deep=True))
        m = map(lambda pb: pb.celldata.fields, cellblocks)
        fields = set(np.concatenate(list(m)))
        for cb in cellblocks:
            id = cb.source().pd.id
            cb.rewire(deep=False, imap=id)
            cb.cd.id = atleast1d(cim.generate_np(len(cb.celldata)))
            for f in fields:
                if f not in cb.celldata.fields:
                    cb.celldata[f] = np.full(len(cb.cd), dcf[f])
            cb.cd.pointdata = None

        # free resources
        for pb in self.pointblocks(inclusive=False, deep=True):
            pb._reset_point_data()

        return self

    def points(
        self, *, return_inds: Optional[bool] = False, from_cells: Optional[bool] = False
    ) -> PointCloud:
        """
        Returns the points as a :class:`~sigmaepsilon.mesh.space.pointcloud.PointCloud` instance.

        Notes
        -----
        Opposed to :func:`coords`, which returns the coordiantes, it returns
        the points of a mesh as vectors.

        See Also
        --------
        :func:`coords`

        Returns
        -------
        :class:`~sigmaepsilon.mesh.space.pointcloud.PointCloud`
        """
        global_frame = self.root.frame

        if from_cells:
            inds_ = np.unique(self.topology())
            x, inds = self.root.points(from_cells=False, return_inds=True)
            imap = inds_to_invmap_as_dict(inds)
            inds = remap_topo_1d(inds_, imap)
            coords, inds = x[inds, :], inds_
        else:
            __cls__ = self.__class__._point_array_class_
            coords, inds = [], []
            for pb in self.pointblocks(inclusive=True):
                x = pb.pd.x
                fr = pb.frame
                i = pb.pd.id
                v = PointCloud(x, frame=fr)
                coords.append(v.show(global_frame))
                inds.append(i)

            if len(coords) == 0:  # pragma: no cover
                raise Exception("There are no points belonging to this block")

            coords = np.vstack(list(coords))
            inds = np.concatenate(inds).astype(int)

        __cls__ = self.__class__._point_array_class_
        points = __cls__(coords, frame=global_frame, inds=inds)

        if return_inds:
            return points, inds
        return points

    def coords(
        self,
        *args,
        return_inds: Optional[bool] = False,
        from_cells: Optional[bool] = False,
        **kwargs,
    ) -> ndarray:
        """
        Returns the coordinates as an array.

        Parameters
        ----------
        return_inds: bool, Optional
            Returns the indices of the points. Default is False.
        from_cells: bool, Optional
            If there is no pointdata attached to the current block, the
            points of the sublevels of the mesh can be gathered from cell
            information. Default is False.

        Returns
        -------
        numpy.ndarray
        """
        if return_inds:
            p, inds = self.points(return_inds=True, from_cells=from_cells)
            return p.show(*args, **kwargs), inds
        else:
            return self.points(from_cells=from_cells).show(*args, **kwargs)

    def bounds(self, *args, **kwargs) -> list:
        """
        Returns the bounds of the mesh.

        Example
        -------
        >>> from sigmaepsilon.mesh.downloads import download_stand
        >>> pd = download_stand(read=True)
        >>> bounds = pd.bounds()
        """
        c = self.coords(*args, **kwargs)
        return [minmax(c[:, 0]), minmax(c[:, 1]), minmax(c[:, 2])]

    def is_2d_mesh(self) -> bool:
        """
        Returns true if the mesh is a 2-dimensional, ie. it only contains 2 dimensional
        cells.
        """
        blocks = self.cellblocks(inclusive=True)
        m = map(lambda b: b.cd.Geometry.number_of_spatial_dimensions, blocks)
        return np.all(np.array(list(m)) == 2)

    def surface_normals(self, *args, **kwargs) -> np.ndarray:
        """
        Retuns the surface normals as a 2d numpy array.

        .. versionadded:: 2.3.0

        Note
        ----
        It only works in cases where the call to `surface` returns a mesh
        with a `normals` method, like a `Trimesh` instance.
        """
        return self.surface(*args, **kwargs).cd.normals()

    def surface_centers(self, *args, **kwargs) -> np.ndarray:
        """
        Retuns the surface centers as a 3d numpy array.

        .. versionadded:: 2.3.0

        Note
        ----
        It only works in cases where the call to `surface` returns a mesh
        with a `normals` method, like a `Trimesh` instance.
        """
        return self.surface(*args, **kwargs).centers()

    @property
    def is_surface(self: PolyDataLike) -> bool:
        blocks: Iterable[PolyData] = list(self.cellblocks(inclusive=True))
        if not len(blocks) == 1:
            return False
        cell_data: PolyCellProtocol = blocks[0].cd
        if not cell_data.Geometry.number_of_spatial_dimensions == 2:
            return False
        return True

    def surface(
        self: PolyDataLike, mesh_class: Optional[Union[PolyDataLike, None]] = None
    ) -> PolyDataLike:
        """
        Returns the surface of the mesh as another `PolyData` instance.

        Parameters
        ----------
        mesh_class: PolyDataLike, Optional
            The class of the resulting mesh instance.
            The default is :class:`sigmaepsilon.mesh.PolyData`.

            .. versionadded:: 2.3.0
        """
        if self.is_surface:
            return self

        if mesh_class is None:
            mesh_class = PolyData

        blocks = list(self.cellblocks(inclusive=True))
        source = self.source()
        coords = source.coords()
        frame = source.frame

        triangles = []
        for block in blocks:
            NDIM = block.celldata.Geometry.number_of_spatial_dimensions
            assert NDIM == 3, "This is only for 3d cells."
            triangles.append(block.cd.extract_surface(detach=False)[-1])
        triangles = np.vstack(triangles)

        if len(blocks) > 1:
            _, indices = np.unique(triangles, axis=0, return_index=True)
            triangles = triangles[indices]

        frames = frames_of_surfaces(coords, triangles)

        pointtype = self.__class__._point_class_
        pd = pointtype(coords=coords, frame=frame)
        cd = Triangle(topo=triangles, pointdata=pd, frames=frames)

        return mesh_class(pd, cd)

    def topology(
        self, *args, return_inds: Optional[bool] = False, **kwargs
    ) -> TopologyArray:
        """
        Returns the topology.

        Parameters
        ----------
        return_inds: bool, Optional
            Returns the indices of the points. Default is False.

        Returns
        -------
        :class:`~sigmaepsilon.mesh.topoarray.TopologyArray`
        """
        blocks = list(self.cellblocks(*args, inclusive=True, **kwargs))
        topo = list(map(lambda i: i.celldata.topology(), blocks))
        topo = np.vstack(topo)

        if return_inds:
            inds = list(map(lambda i: i.celldata.id, blocks))
            return topo, np.concatenate(inds)
        else:
            return topo

    def cell_indices(self) -> ndarray:
        """
        Returns the indices of the cells along the walk.
        """
        blocks = self.cellblocks(inclusive=True)
        m = map(lambda b: b.cd.id, blocks)
        return np.concatenate(list(m))

    def detach(self: PolyDataLike, nummrg: Optional[bool] = False) -> PolyDataLike:
        """
        Returns a detached version of the mesh.

        Parameters
        ----------
        nummrg: bool, Optional
            If True, merges node numbering. Default is False.
        """
        s: PolyData = self.source()
        polydata = PolyData(s.pd, frame=s.frame)
        l0 = len(self.address)

        if self.celldata is not None:
            db = deepcopy(self.cd.db)
            cd = self.celltype(container=polydata, db=db)
            polydata.celldata = cd
            polydata.celltype = self.celltype

        for cb in self.cellblocks(inclusive=False):
            addr = cb.address
            if len(addr) > l0:
                db = deepcopy(cb.cd.db)
                cd = cb.celltype(container=polydata, db=db)
                assert cd is not None
                polydata[addr[l0:]] = PolyData(None, cd)
                assert polydata[addr[l0:]].celldata is not None

        if nummrg:
            polydata.nummrg()

        return polydata

    def nummrg(self: PolyDataLike) -> PolyDataLike:
        """
        Merges node numbering.
        """
        assert self.is_root(), "This must be called on he root object!"
        topo = self.topology()
        inds = np.unique(topo)
        pointtype = self.__class__._point_class_
        self.pointdata = pointtype(db=self.pd[inds])
        imap = inds_to_invmap_as_dict(self.pd.id)
        [cb.rewire(imap=imap) for cb in self.cellblocks(inclusive=True)]
        self.pointdata.id = np.arange(len(self.pd))
        return self

    def move(
        self: PolyDataLike,
        v: VectorLike,
        frame: Optional[Union[FrameLike, None]] = None,
        inplace: Optional[bool] = True,
    ) -> PolyDataLike:
        """
        Moves and returns the object or a deep copy of it.

        Parameters
        ----------
        v: VectorLike, Optional
            A vector describing a translation.
        frame: :class:`~sigmaepsilon.math.linalg.FrameLike`, Optional
            If `v` is only an array, this can be used to specify
            a frame in which the components should be understood.
        inplace: bool, Optional
            If True, the transformation is done on the instance, otherwise
            a deep copy is created first. Default is True.

        Examples
        --------
        Download the Stanford bunny and move it along global X:

        >>> from sigmaepsilon.mesh.downloads import download_bunny
        >>> import numpy as np
        >>> bunny = download_bunny(tetra=False, read=True)
        >>> bunny.move([0.2, 0, 0])
        PolyData({5: PolyData({})})
        """
        subject = self if inplace else self.deepcopy()
        if subject.is_source():
            pc = subject.points()
            pc.move(v, frame)
            subject.pointdata.x = pc.array
        else:  # pragma: no cover
            raise Exception("This is only for blocks with a point source.")
        return subject

    def rotate(
        self: PolyDataLike, *args, inplace: Optional[bool] = True, **kwargs
    ) -> PolyDataLike:
        """
        Rotates and returns the object. Positional and keyword arguments
        not listed here are forwarded to :class:`sigmaepsilon.math.linalg.frame.ReferenceFrame`

        Parameters
        ----------
        *args
            Forwarded to :class:`sigmaepsilon.math.linalg.frame.ReferenceFrame`.
        inplace: bool, Optional
            If True, the transformation is done on the instance, otherwise
            a deep copy is created first. Default is True.
        **kwargs
            Forwarded to :class:`sigmaepsilon.math.linalg.frame.ReferenceFrame`.

        Examples
        --------
        Download the Stanford bunny and rotate it about global Z with 90 degrees:

        >>> from sigmaepsilon.mesh.downloads import download_bunny
        >>> import numpy as np
        >>> bunny = download_bunny(tetra=False, read=True)
        >>> bunny.rotate("Space", [0, 0, np.pi/2], "xyz")
        PolyData({5: PolyData({})})
        """
        subject = self if inplace else self.deepcopy()

        if subject.is_source():
            pc = subject.points()
            source = subject
        else:  # pragma: no cover
            raise Exception("This is only for blocks with a point source.")

        pc.rotate(*args, **kwargs)
        subject._rotate_attached_cells_(*args, **kwargs)
        source.pointdata.x = pc.show(subject.frame)

        return subject

    def spin(
        self: PolyDataLike, *args, inplace: Optional[bool] = True, **kwargs
    ) -> PolyDataLike:
        """
        Like rotate, but rotation happens around centroidal axes. Positional and keyword
        arguments not listed here are forwarded to :class:`sigmaepsilon.math.linalg.frame.ReferenceFrame`

        Parameters
        ----------
        *args
            Forwarded to :class:`sigmaepsilon.math.linalg.frame.ReferenceFrame`.
        inplace: bool, Optional
            If True, the transformation is done on the instance, otherwise
            a deep copy is created first. Default is True.
        **kwargs
            Forwarded to :class:`sigmaepsilon.math.linalg.frame.ReferenceFrame`.

        Examples
        --------
        Download the Stanford bunny and spin it about global Z with 90 degrees:

        >>> from sigmaepsilon.mesh.downloads import download_bunny
        >>> import numpy as np
        >>> bunny = download_bunny(tetra=False, read=True)
        >>> bunny.spin("Space", [0, 0, np.pi/2], "xyz")
        PolyData({5: PolyData({})})
        """
        subject = self if inplace else self.deepcopy()

        if subject.is_source():
            pc = subject.points()
            source = subject
        else:  # pragma: no cover
            raise Exception("This is only for blocks with a point source.")

        center = pc.center()
        pc.centralize()
        pc.rotate(*args, **kwargs)
        pc.move(center)
        subject._rotate_attached_cells_(*args, **kwargs)
        source.pointdata.x = pc.show(subject.frame)

        return subject

    def cells_at_nodes(self, *args, **kwargs) -> Iterable:
        """
        Returns the neighbouring cells of nodes.

        Returns
        -------
        object
            Some kind of iterable, depending on the inputs.
            See the docs below for further details.

        See Also
        --------
        :func:`cells_at_nodes`
        """
        topo = self.topology()

        if isinstance(topo, TopologyArray):
            if topo.is_jagged():
                topo = topo.to_csr()
            else:
                topo = topo.to_numpy()

        return cells_at_nodes(topo, *args, **kwargs)

    def cells_around_cells(
        self, radius: float, frmt: str = "dict"
    ) -> Union[JaggedArray, csr_matrix, dict]:
        """
        Returns the neares cells to cells.

        Parameters
        ----------
        radius: float
            The influence radius of a point.
        frmt: str, Optional
            A string specifying the type of the result. Valid
            options are 'jagged', 'csr' and 'dict'.

        See Also
        --------
        :func:`cells_around`
        """
        return cells_around(self.centers(), radius, frmt=frmt)

    def nodal_adjacency(self, *args, **kwargs) -> Any:
        """
        Returns the nodal adjecency matrix.

        Parameters
        ----------
        frmt: str
            A string specifying the output format. Valid options are
            'jagged', 'csr', 'nx' and 'scipy-csr'. See below for the details on the
            returned object.
        assume_regular: bool
            If the topology is regular, you can gain some speed with providing
            it as `True`. Default is `False`.
        """
        # FIXME This doesn't work with Awkward arrays.
        # topo = self.topology(jagged=True).to_array()
        topo = self.topology(jagged=True).to_numpy()

        if isinstance(topo, ak.Array):
            topo = ak.values_astype(topo, "int64")
        else:
            assert isinstance(topo, ndarray)
            topo = topo.astype(np.int64)

        return nodal_adjacency(topo, *args, **kwargs)

    def nodal_adjacency_matrix(
        self, assume_regular: Optional[bool] = False
    ) -> spmatrix:
        """
        Returns the nodal adjecency information as a SciPy CSR matrix.

        Parameters
        ----------
        assume_regular: bool
            If the topology is regular, you can gain some speed with providing
            it as `True`. Default is `False`.

            .. versionadded:: 2.3.0

        Returns
        -------
        scipy.sparse.spmatrix
        """
        return self.nodal_adjacency(frmt="scipy-csr", assume_regular=assume_regular)

    def nodal_neighbourhood_matrix(self) -> csr_scipy:
        """
        Returns a sparse SciPy CSR matrix as a representation of the first order
        neighbourhood structure of the mesh.

        The [i, j] entry of the returned matrix is 1 if points i and j are
        neighbours (they share a cell) 0 if they are not. Points are not considered
        to be neighbours of themselfes, therefore entries in the main diagonal are zero.

        .. versionadded:: 2.3.0
        """
        adj: spmatrix = self.nodal_adjacency_matrix()
        adj_csr = csr_scipy(adj - scipy_diags(adj.diagonal()))
        adj_csr.sum_duplicates()
        adj_csr.eliminate_zeros()
        return adj_csr

    def number_of_cells(self) -> int:
        """Returns the number of cells."""
        blocks = self.cellblocks(inclusive=True)
        return np.sum(list(map(lambda i: len(i.celldata), blocks)))

    def number_of_points(self) -> int:
        """Returns the number of points."""
        return len(self.source().pointdata)

    def cells_coords(self) -> ndarray:
        """Returns the coordiantes of the cells in explicit format."""
        return cells_coords(self.source().coords(), self.topology().to_numpy())

    def center(self, target: Optional[Union[FrameLike, None]] = None) -> ndarray:
        """
        Returns the center of the pointcloud of the mesh.

        Parameters
        ----------
        target: :class:`~sigmaepsilon.math.linalg.FrameLike`, Optional
            The target frame in which the returned coordinates are to be understood.
            A `None` value means the frame the mesh is embedded in. Default is None.

        Returns
        -------
        numpy.ndarray
            A one dimensional float array.
        """
        centers = self.centers(target)
        return np.array(
            [np.mean(centers[:, i]) for i in range(centers.shape[1])],
            dtype=centers.dtype,
        )

    def centers(self, target: Optional[Union[FrameLike, None]] = None) -> ndarray:
        """
        Returns the centers of the cells.

        Parameters
        ----------
        target: :class:`~sigmaepsilon.math.linalg.FrameLike`, Optional
            The target frame in which the returned coordinates are to be understood.
            A `None` value means the frame the mesh is embedded in. Default is None.

        Returns
        -------
        numpy.ndarray
            A 2 dimensional float array.
        """
        source = self.source()
        coords = source.coords()
        blocks = self.cellblocks(inclusive=True)

        def foo(b: PolyData[PointData, PolyCell]):
            t = b.cd.topology().to_numpy()
            return cell_centers_bulk(coords, t)

        centers = np.vstack(list(map(foo, blocks)))

        if target:
            pc = PointCloud(centers, frame=source.frame)
            centers = pc.show(target)

        return centers

    def centralize(
        self: PolyDataLike,
        target: Optional[Union[FrameLike, None]] = None,
        inplace: Optional[bool] = True,
        axes: Optional[Union[Iterable[int], None]] = None,
    ) -> PolyDataLike:
        """
        Moves all the meshes that belong to the same source such that the current object's
        center will be at the origin of its embedding frame.

        Parameters
        ----------
        target: :class:`~sigmaepsilon.math.linalg.FrameLike`, Optional
            The target frame the mesh should be central to. A `None` value
            means the frame the mesh is embedded in. Default is True.
        inplace: bool, Optional
            If True, the transformation is done on the instance, otherwise
            a deep copy is created first. Default is True.
        axes: Iterable[int], Optional
            The axes on which centralization is to be performed. A `None` value
            means all axes. For instance providing `axes=[2]` would only centralize
            coordinates in Z direction. Default is None.

        Notes
        -----
        This operation changes the coordinates of all blocks that belong to the same
        pointcloud as the object the function is called on.
        """
        subject = self if inplace else self.deepcopy()
        source = subject.source()
        target = source.frame if target is None else target
        center = self.center(target)
        if axes is not None:
            all_axes = set([0, 1, 2])
            input_axes = set(axes)
            missing_axes = list(all_axes - input_axes)
            center[missing_axes] = 0.0
        for block in source.pointblocks(inclusive=True):
            block_points = block.pd.x
            block.pd.x = block_points - center
        return subject

    def k_nearest_cell_neighbours(
        self, k, *args, knn_options: Optional[Union[dict, None]] = None, **kwargs
    ):
        """
        Returns the k closest neighbours of the cells of the mesh, based
        on the centers of each cell.

        The argument `knn_options` is passed to the KNN search algorithm,
        the rest to the `centers` function of the mesh.

        Examples
        --------
        >>> from sigmaepsilon.mesh.grid import grid
        >>> from sigmaepsilon.mesh import KNN
        >>> size = 80, 60, 20
        >>> shape = 10, 8, 4
        >>> X, _ = grid(size=size, shape=shape, eshape='H8')
        >>> i = KNN(X, X, k=3, max_distance=10.0)

        See Also
        --------
        :func:`KNN`
        """
        c = self.centers(*args, **kwargs)
        knn_options = {} if knn_options is None else knn_options
        return KNN(c, c, k=k, **knn_options)

    def areas(self, *args, **kwargs) -> ndarray:
        """Returns the areas."""
        blocks = self.cellblocks(*args, inclusive=True, **kwargs)
        blocks2d = filter(
            lambda b: b.celltype.Geometry.number_of_spatial_dimensions < 3, blocks
        )
        amap = map(lambda b: b.celldata.areas(), blocks2d)
        return np.concatenate(list(amap))

    def area(self, *args, **kwargs) -> float:
        """Returns the sum of areas in the model."""
        return np.sum(self.areas(*args, **kwargs))

    def volumes(self, *args, **kwargs) -> ndarray:
        """Returns the volumes of the cells."""
        blocks = self.cellblocks(*args, inclusive=True, **kwargs)
        vmap = map(lambda b: b.celldata.volumes(), blocks)
        return np.concatenate(list(vmap))

    def volume(self, *args, **kwargs) -> float:
        """Returns the net volume of the mesh."""
        return np.sum(self.volumes(*args, **kwargs))

    def index_of_closest_point(self, target: Iterable) -> int:
        """Returns the index of the closest point to a target."""
        return index_of_closest_point(self.coords(), np.array(target, dtype=float))

    def index_of_furthest_point(self, target: Iterable) -> int:
        """
        Returns the index of the furthest point to a target.
        """
        return index_of_furthest_point(self.coords(), np.array(target, dtype=float))

    def index_of_closest_cell(self, target: Iterable) -> int:
        """Returns the index of the closest cell to a target."""
        return index_of_closest_point(self.centers(), np.array(target, dtype=float))

    def index_of_furthest_cell(self, target: Iterable) -> int:
        """
        Returns the index of the furthest cell to a target.
        """
        return index_of_furthest_point(self.centers(), np.array(target, dtype=float))

    def nodal_distribution_factors(
        self, weights: Optional[Union[str, ndarray]] = "volume"
    ) -> Union[ndarray, csr_matrix]:
        """
        Retruns nodal distribution factors for all nodes of all cells
        as a 2d array. The returned array has the same shape as the
        topology array, where the j-th factor of the i-th row is the
        contribution of element i to the j-th node of the cell.

        Parameters
        ----------
        weights: Union[str, numpy.ndarray], Optional
            The metric which is used to calculate the factors. Valid
            strings are 'volume' and 'uniform'. If it is an array, it
            must be an 1d array with a length matching the number of
            cells. Default is 'volume'.

        Returns
        -------
        numpy.ndarray or sigmaepsilon.math.linalg.sparse.csr.csr_matrix
            An array with the same shape as the topology.

        Note
        ----
        For a given node, the sum of all contribution factors from all
        the cells that meet at that node is one.

        See also
        --------
        :func:`~sigmaepsilon.mesh.utils.utils.nodal_distribution_factors`
        """
        assert self.is_source(), "This can only be called on objects with PointData."
        topo = self.topology()

        if isinstance(topo, TopologyArray):
            if topo.is_jagged():
                topo = topo.to_csr()
            else:
                topo = topo.to_numpy()

        if isinstance(weights, str):
            if weights == "volume":
                weights = self.volumes()
            elif weights == "uniform":
                weights = np.ones(topo.shape[0], dtype=float)

        assert isinstance(weights, ndarray), "'weights' must be a NumPy array!"
        assert len(weights) == topo.shape[0], (
            "Mismatch in shape. The weights must have the same number of "
            + "values as cells in the block."
        )

        return nodal_distribution_factors(topo, weights)

    def _rotate_attached_cells_(self, *args, **kwargs):
        for block in self.cellblocks(inclusive=True):
            block.cd._rotate_(*args, **kwargs)

    def _in_all_pointdata_(self, key: str) -> bool:
        blocks = self.pointblocks(inclusive=True)
        return all(list(map(lambda b: key in b.pointdata.db.fields, blocks)))

    def _in_all_celldata_(self, key: str) -> bool:
        blocks = self.cellblocks(inclusive=True)
        return all(list(map(lambda b: key in b.celldata.db.fields, blocks)))

    def _detach_block_data_(self, data: Union[str, ndarray] = None) -> Iterator:
        blocks = self.cellblocks(inclusive=True, deep=True)
        for block in blocks:
            source = block.source()
            coords = source.coords()
            topo = block.topology().to_numpy()

            point_data = None
            if isinstance(data, ndarray):
                if not data.shape[0] == len(source.pd):
                    raise ValueError(
                        "The length of scalars must match the number of points."
                    )
                point_data = data
            elif isinstance(data, str):
                if data in source.pd.fields:
                    point_data = source.pd.db[data].to_numpy()
            else:
                if data is not None:
                    if not isinstance(data, str):
                        raise TypeError("Data must be a NumPy array or a string.")

            if point_data is not None:
                c, d, t = detach_mesh_data_bulk(coords, topo, point_data)
                yield block, c, t, d
            else:
                c, t = detach_mesh_bulk(coords, topo)
                if data is not None:
                    if data in block.cd.fields:
                        d = block.cd.db[data].to_numpy()
                        if len(d.shape) == 2:
                            c, t, d = explode_mesh_data_bulk(c, t, d)
                        else:
                            assert len(d.shape) == 1, "Cell data must be 1d or 2d."
                        yield block, c, t, d
                    else:
                        yield block, c, t, None
                else:
                    yield block, c, t, None

    def _has_plot_scalars_(self, scalars: Union[str, ndarray]):
        """
        Returns a boolean value for every cell block in the mesh. A value
        for a block is True, if data is provided for plotting. If 'scalars'
        is a NumPy array, it is assumed that the mesh is centralized an therefore
        all values are True. Otherwise the data key must be a sting and if data is found
        in a blocks database or in the database of the related source, the value is True.
        """
        res = []
        for block in self.cellblocks(inclusive=True, deep=True):
            if isinstance(scalars, ndarray):
                res.append(True)
            elif isinstance(scalars, str):
                if block.source(scalars) is not None:
                    res.append(True)
                elif scalars in block.cd.fields:
                    res.append(True)
                else:
                    res.append(False)
            elif scalars is None:
                res.append(False)
            else:
                raise ValueError("'scalars' must be a string or a NumPy array")
        return res

    def _get_config_(self, key: str) -> dict:
        if key in self.config:
            return self.config[key]
        else:
            if self.parent is not None:
                return self.parent._get_config_(key)
            else:
                return {}

    if __hasvtk__:

        def to_vtk(
            self, deepcopy: Optional[bool] = False, multiblock: Optional[bool] = False
        ) -> Union[vtk.vtkUnstructuredGrid, vtk.vtkMultiBlockDataSet]:
            """
            Returns the mesh as a `VTK` object.

            Parameters
            ----------
            deepcopy: bool, Optional
                Default is False.
            multiblock: bool, Optional
                Wether to return the blocks as a `vtkMultiBlockDataSet` or a list
                of `vtkUnstructuredGrid` instances. Default is False.

            Returns
            -------
            vtk.vtkUnstructuredGrid or vtk.vtkMultiBlockDataSet
            """
            exporter: Callable = exporters["vtk"]
            return exporter(self, deepcopy=deepcopy, multiblock=multiblock)

    if __hasvtk__ and __haspyvista__:

        def to_pv(
            self,
            deepcopy: Optional[bool] = False,
            multiblock: Optional[bool] = False,
            scalars: Optional[Union[str, ndarray, None]] = None,
        ) -> Union[pv.UnstructuredGrid, pv.MultiBlock]:
            """
            Returns the mesh as a `PyVista` object, optionally set up with data.

            Parameters
            ----------
            deepcopy: bool, Optional
                Default is False.
            multiblock: bool, Optional
                Wether to return the blocks as a `vtkMultiBlockDataSet` or a list
                of `vtkUnstructuredGrid` instances. Default is False.
            scalars: str or numpy.ndarray, Optional
                A string or an array describing scalar data. Default is None.

            Returns
            -------
            pyvista.UnstructuredGrid or pyvista.MultiBlock
            """
            exporter: Callable = exporters["PyVista"]
            return exporter(
                self, deepcopy=deepcopy, multiblock=multiblock, scalars=scalars
            )

    if __hask3d__:

        def to_k3d(self, *args, **kwargs) -> k3d.Plot:
            """
            Returns the mesh as a k3d mesh object. All arguments are forwarded to
            :func:~`sigmaepsilon.mesh.io.to_k3d.to_k3d`, refer to its documentation
            for the details.

            :: warning:
                Calling this method raises a UserWarning inside the `traittypes`
                package saying "Given trait value dtype 'float32' does not match
                required type 'float32'." However, plotting seems to be fine.

            Returns
            -------
            k3d.Plot
                A K3D Plot Widget, which is a result of a call to `k3d.plot`.
            """
            exporter: Callable = exporters["k3d"]
            return exporter(self, *args, **kwargs)

        def k3dplot(self, *args, **kwargs) -> k3d.Plot:
            """
            Convenience function for plotting the mesh using K3D. All arguments are
            forwarded to :func:~`sigmaepsilon.mesh.plotting.k3dplot.k3dplot`, refer the
            documentation of this function for the details.

            .. warning::
                During this call there is a UserWarning saying 'Given trait value dtype
                "float32" does not match required type "float32"'. Although this is weird,
                plotting seems to be just fine.

            Returns
            -------
            k3d.Plot
                A K3D Plot Widget, which is a result of a call to `k3d.plot`.

            See Also
            --------
            :func:`to_k3d`
            :func:`k3d.plot`
            """
            plotter: Callable = plotters["k3d"]
            return plotter(self, *args, **kwargs)

    if __haspyvista__:

        def pvplot(self, *args, **kwargs) -> Union[None, pv.Plotter, np.ndarray]:
            """
            Convenience function for plotting the mesh using PyVista. All arguments are
            forwarded to :func:~`sigmaepsilon.mesh.plotting.pvplot.pvplot`, refer the
            documentation of this function for the details.

            .. note::
                See https://github.com/pyvista/pyvista for more examples with
                ``pyvista``. Be sure to install ``pyvista`` with ``pip install
                pyvista`` if you wish to use it.

            Returns
            -------
            Union[None, pv.Plotter, numpy.ndarray]
                A PyVista plotter if `return_plotter` is `True`, a NumPy array if
                `return_img` is `True`, or nothing.

            See Also
            --------
            :func:`to_pv`
            :func:`to_vtk`
            """
            plotter: Callable = plotters["PyVista"]
            return plotter(self, *args, **kwargs)

    def plot(
        self,
        *,
        notebook: Optional[bool] = False,
        backend: Optional[str] = "pyvista",
        config_key: Optional[Union[str, None]] = None,
        **kwargs,
    ) -> Any:
        """
        Plots the mesh using supported backends. The default backend is PyVista.

        Parameters
        ----------
        notebook: bool, Optional
            Whether to plot in an IPython notebook enviroment. This is only
            available for PyVista at the moment. Default is False.
        backend: str, Optional
            The backend to use. Valid options are 'k3d' and 'pyvista'.
            Default is 'pyvista'.
        config_key: str, Optional
            A configuration key if the block were configured previously.
            Default is None.
        **kwargs: dict, Optional
            Extra keyword arguments forwarded to the plotter function according
            to the selected backend.

        See Also
        --------
        :func:`pvplot`
        :func:`k3dplot`
        """
        backend = backend.lower()
        if notebook and backend == "k3d":
            return self.k3dplot(config_key=config_key, **kwargs)
        elif backend == "pyvista":
            return self.pvplot(notebook=notebook, config_key=config_key, **kwargs)

    def __join_parent__(self, parent: DeepDict, key: Hashable = None) -> None:
        super().__join_parent__(parent, key)
        if self.celldata is not None:
            GIDs = self.root.cim.generate_np(len(self.celldata.db))
            self.celldata.db.id = atleast1d(GIDs)
            if self.celldata.pd is None:
                self.celldata.pd = self.source().pd
            self.celldata.container = self

    def __leave_parent__(self) -> None:
        if self.celldata is not None:
            self.root.cim.recycle(self.celldata.db.id)
            dbkey = self.celldata.db._dbkey_id_
            del self.celldata.db._wrapped[dbkey]
        super().__leave_parent__()

    def __repr__(self):
        return "PolyData(%s)" % (dict.__repr__(self))



# -*- coding: utf-8 -*-
import numpy as np

from .config import __hasvtk__

if __hasvtk__:
    import vtk
    from vtk.util.numpy_support import (
        numpy_to_vtk as np2vtk,
        numpy_to_vtkIdTypeArray as np2vtkId,
    )
    from vtk.numpy_interface import dataset_adapter as dsa


def mesh_to_vtkdata(coords, topo, deepcopy=True):
    if not __hasvtk__:
        raise ImportError
    # points
    vtkpoints = vtk.vtkPoints()
    vtkpoints.SetData(np2vtk(coords, deep=deepcopy))
    # cells
    topo_vtk = np.concatenate(
        (np.ones((topo.shape[0], 1), dtype=int) * topo.shape[1], topo), axis=1
    ).ravel()
    vtkcells = vtk.vtkCellArray()
    vtkcells.SetNumberOfCells(topo.shape[0])
    vtkcells.SetCells(topo.shape[0], np2vtkId(topo_vtk, deep=deepcopy))
    # return points and cells
    return vtkpoints, vtkcells


def mesh_to_UnstructuredGrid(coords, topo, vtkCellType, deepcopy=True):
    vtkpoints, vtkcells = mesh_to_vtkdata(
        coords.astype(np.float64), topo.astype(np.int64), deepcopy
    )
    ugrid = vtk.vtkUnstructuredGrid()
    ugrid.SetPoints(vtkpoints)
    ugrid.SetCells(vtkCellType, vtkcells)
    return ugrid


def mesh_to_PolyData(coords, topo, deepcopy=True):
    vtkpoints, vtkcells = mesh_to_vtkdata(coords, topo, deepcopy)
    vtkPolyData = vtk.vtkPolyData()
    vtkPolyData.SetPoints(vtkpoints)
    vtkPolyData.SetPolys(vtkcells)
    vtkPolyData.Modified()
    return vtkPolyData


def PolyData_to_mesh(pd, v=0):
    if v == 0:
        coords = pd.points.astype(float)
        ugrid = pd.cast_to_unstructured_grid()
        topo = ugrid.cells_dict
    elif v == 1:
        with dsa.WrapDataObject(pd) as wrapper:
            coords = wrapper.Points
            topo = wrapper.Polygons
    elif v == 2:
        coords = vtk_to_numpy(pd.GetPoints().GetData())
        topo = vtk_to_numpy(pd.GetPolys().GetData())
    return coords, topo



import numpy as np
from numpy import ndarray
from numba import njit, prange

__cache = True


@njit(nogil=True, cache=__cache)
def vol_tet(ecoords: ndarray) -> ndarray:
    """
    Calculates volume of a single tetrahedron.

    Parameters
    ----------
    ecoords: numpy.ndarray
        A 3d float array of shape (nNE=4, 3) of nodal coordinates.

    Returns
    -------
    float
        The volume of the tetrahedron.

    Notes
    -----
    1) This only returns exact results for linear cells. For
    nonlinear cells, use objects that calculate the volumes
    using numerical integration.
    2) This function is numba-jittable in 'nopython' mode.
    """
    v1 = ecoords[1] - ecoords[0]
    v2 = ecoords[2] - ecoords[0]
    v3 = ecoords[3] - ecoords[0]
    return np.dot(np.cross(v1, v2), v3) / 6


@njit(nogil=True, parallel=True, cache=__cache)
def vol_tet_bulk(ecoords: ndarray) -> ndarray:
    """
    Calculates volumes of several tetrahedra.

    Parameters
    ----------
    ecoords: numpy.ndarray
        A 3d float array of shape (nE, nNE=4, 3) of
        nodal coordinates for several elements. Here nE
        is the number of nodes and nNE is the number of
        nodes per element.

    Returns
    -------
    numpy.ndarray
        1d float array of volumes.

    Notes
    -----
    1) This only returns exact results for linear cells. For
    nonlinear cells, use objects that calculate the volumes
    using numerical integration.
    2) This function is numba-jittable in 'nopython' mode.
    """
    nE = len(ecoords)
    res = np.zeros(nE, dtype=ecoords.dtype)
    for i in prange(nE):
        res[i] = vol_tet(ecoords[i])
    return np.abs(res)


@njit(nogil=True, cache=__cache)
def glob_to_nat_tet(gcoord: ndarray, ecoords: ndarray) -> ndarray:
    """
    Transformation from global to natural coordinates within a tetrahedron
    for a single point and tetrahedra.

    Notes
    -----
    This function is numba-jittable in 'nopython' mode.
    """
    ecoords_ = np.zeros_like(ecoords)
    ecoords_[:, :] = ecoords[:, :]
    V = vol_tet(ecoords)

    ecoords_[0, :] = gcoord
    v1 = vol_tet(ecoords_) / V
    ecoords_[0, :] = ecoords[0, :]

    ecoords_[1, :] = gcoord
    v2 = vol_tet(ecoords_) / V
    ecoords_[1, :] = ecoords[1, :]

    ecoords_[2, :] = gcoord
    v3 = vol_tet(ecoords_) / V
    ecoords_[2, :] = ecoords[2, :]

    return np.array([v1, v2, v3, 1 - v1 - v2 - v3], dtype=gcoord.dtype)


@njit(nogil=True, parallel=True, cache=__cache)
def _glob_to_nat_tet_bulk_(points: ndarray, ecoords: ndarray) -> ndarray:
    nE = ecoords.shape[0]
    nP = points.shape[0]
    res = np.zeros((nP, nE, 4), dtype=points.dtype)
    for i in prange(nP):
        for j in prange(nE):
            res[i, j, :] = glob_to_nat_tet(points[i], ecoords[j])
    return res


@njit(nogil=True, cache=__cache)
def pip_tet(point: ndarray, ecoords: ndarray, tol: float = 1e-12) -> ndarray:
    """
    Tells if a point is inside a tetrahedron or not.

    Parameters
    ----------
    point: numpy.ndarray
        1d NumPy array of the global coordinates of the point.
    ecoords: numpy.ndarray
        2d NumPy array of the coordinates of the nodes of the tetrahedron.
    tol: float, Optional
        Tolerance to consider a point inside a cell. Default is 1e-12.

    Returns
    -------
    bool
        True if the point is inside the tetrahedron, False otherwise.

    Notes
    -----
    This function is numba-jittable in 'nopython' mode.
    """
    nat = glob_to_nat_tet(point, ecoords)
    return np.all(nat > -tol) and np.all(nat < (1 + tol))


@njit(nogil=True, parallel=True, cache=__cache)
def _pip_tet_bulk_nat_(nat: ndarray, tol: float = 1e-12) -> ndarray:
    nP, nE = nat.shape[:2]
    res = np.zeros((nP, nE), dtype=np.bool_)
    for i in prange(nP):
        for j in prange(nE):
            c1 = np.all(nat[i, j] > (-tol))
            c2 = np.all(nat[i, j] < (1 + tol))
            res[i, j] = c1 & c2
    return res


@njit(nogil=True, cache=__cache)
def _pip_tet_bulk_(points: ndarray, ecoords: ndarray, tol: float = 1e-12) -> ndarray:
    nat = _glob_to_nat_tet_bulk_(points, ecoords)
    return _pip_tet_bulk_nat_(nat, tol)


@njit(nogil=True, cache=__cache)
def _pip_tet_bulk_knn_(
    points: ndarray, ecoords: ndarray, neighbours: ndarray, tol: float = 1e-12
) -> ndarray:
    nat = _glob_to_nat_tet_bulk_knn_(points, ecoords, neighbours)
    return _pip_tet_bulk_nat_(nat, tol)


def pip_tet_bulk(
    points: ndarray,
    ecoords: ndarray,
    tol: float = 1e-12,
    neighbours: ndarray | None = None,
) -> ndarray:
    """
    Tells if points are inside tetrahedra or not.

    Parameters
    ----------
    points: numpy.ndarray
        2d NumPy array of the global coordinates of the points.
    ecoords: numpy.ndarray
        3d NumPy array of the coordinates of the nodes of the tetrahedra.
    tol: float, Optional
        Tolerance to consider a point inside a cell. Default is 1e-12.
    neighbours: numpy.ndarray, Optional
        2d NumPy array of the indices of the neighbouring cells of each point.
        These indices refer to the tetrahedra in the `ecoords` array. Providing
        this is optional, but strogly recommended for large meshes.
    """
    if neighbours is None:
        return _pip_tet_bulk_(points, ecoords, tol)
    else:
        return _pip_tet_bulk_knn_(points, ecoords, neighbours, tol)


@njit(nogil=True, parallel=True, cache=__cache)
def _glob_to_nat_tet_bulk_knn_(
    points: ndarray, ecoords: ndarray, neighbours: ndarray
) -> ndarray:
    kE = neighbours.shape[1]
    nP = points.shape[0]
    res = np.zeros((nP, kE, 4), dtype=points.dtype)
    for i in prange(nP):
        for k in prange(kE):
            res[i, k, :] = glob_to_nat_tet(points[i], ecoords[neighbours[i, k]])
    return res


@njit(nogil=True, cache=__cache)
def lcoords_tet(center: ndarray = None) -> ndarray:
    """
    Returns coordinates of the master element
    of a simplex in 3d.

    Parameters
    ----------
    center: numpy.ndarray, Optional
        The coordinates of the center of the master
        element. Default is None.

    Notes
    -----
    This function is numba-jittable in 'nopython' mode.
    """
    res = np.array(
        [
            [-1 / 3, -1 / 3, -1 / 3],
            [2 / 3, -1 / 3, -1 / 3],
            [-1 / 3, 2 / 3, -1 / 3],
            [-1 / 3, -1 / 3, 2 / 3],
        ]
    )
    if center is not None:
        res += center
    return res


@njit(nogil=True, cache=__cache)
def nat_to_loc_tet(
    acoord: ndarray, lcoords: ndarray | None = None, center: ndarray | None = None
) -> ndarray:
    """
    Transformation from natural to local coordinates
    within a tetrahedra.

    Parameters
    ----------
    acoord: numpy.ndarray
        1d NumPy array of natural coordinates of a point.
    lcoords: numpy.ndarray, Optional
        2d NumPy array of parametric coordinates (r, s, t) of the
        master cell of a tetrahedron.
    center: numpy.ndarray
        The local coordinates (r, s, t) of the geometric center
        of the master tetrahedron. If not provided it is assumed to
        be at (0, 0, 0).

    Notes
    -----
    This function is numba-jittable in 'nopython' mode.
    """
    if lcoords is None:
        lcoords = lcoords_tet(center)
    return acoord.T @ lcoords



from meshio._vtk_common import vtk_to_meshio_type

from .abcakwrapper import ABCMeta_AkWrapper
from .geometry import GeometryProtocol
from ..helpers import vtk_to_celltype, meshio_to_celltype


__all__ = ["ABCMeta_PolyCell", "ABC_PolyCell"]


class ABCMeta_PolyCell(ABCMeta_AkWrapper):
    """
    Meta class for PointData and CellData classes.

    It merges attribute maps with those of the parent classes.
    """

    def __init__(self, name, bases, namespace, *args, **kwargs):
        super().__init__(name, bases, namespace, *args, **kwargs)

    def __new__(metaclass, name, bases, namespace, *args, **kwargs):
        cls = super().__new__(metaclass, name, bases, namespace, *args, **kwargs)

        if namespace.get("Geometry", None):
            if not isinstance(cls.Geometry, GeometryProtocol):
                raise TypeError(
                    f"The attached geometry class {cls.Geometry} of {cls} "
                    "does not implement PolyCellGeometry"
                )
            else:
                # add class to helpers
                vtk_cell_id = getattr(cls.Geometry, "vtk_cell_id", None)
                if isinstance(vtk_cell_id, int):
                    vtk_to_celltype[vtk_cell_id] = cls
                    meshio_to_celltype[vtk_to_meshio_type[vtk_cell_id]] = cls

        return cls


class ABC_PolyCell(metaclass=ABCMeta_PolyCell):
    """
    Helper class that provides a standard way to create an ABC using
    inheritance.
    """

    __slots__ = ()



from numba import njit
import numpy as np
from numpy import ndarray

__cache = True


@njit(nogil=True, cache=__cache)
def monoms_L3(r: float) -> ndarray:
    return np.array([1, r, r**2], dtype=float)



from ..config import __haspyvista__
from ..helpers import plotters

if not __haspyvista__:  # pragma: no cover

    def pvplot(*_, **__) -> None:
        raise ImportError(
            "You need PyVista for this. Install it with 'pip install pyvista'. "
            "You may also need to restart your kernel and reload the package."
        )

else:
    from typing import Union, Iterable, Tuple
    from copy import copy

    import pyvista as pv
    from pyvista import themes
    from numpy import ndarray

    from ..data import PolyData

    def pvplot(
        obj: PolyData,
        *,
        deepcopy: bool = False,
        jupyter_backend: str = "pythreejs",
        show_edges: bool = True,
        notebook: bool = False,
        theme: str = None,
        scalars: Union[str, ndarray] = None,
        window_size: Tuple = None,
        return_plotter: bool = False,
        config_key: Tuple = None,
        plotter: pv.Plotter = None,
        cmap: Union[str, Iterable] = None,
        camera_position: Tuple = None,
        lighting: bool = False,
        edge_color: str = None,
        return_img: bool = False,
        show_scalar_bar: Union[bool, None] = None,
        **kwargs,
    ) -> Union[None, pv.Plotter, ndarray]:
        """
        Plots the mesh using PyVista. The parameters listed here only grasp
        a fraction of what PyVista provides. The idea is to have a function
        that narrows down the parameters as much as possible to the ones that
        are most commonly used. If you want more control, create a plotter
        prior to calling this function and provide it using the parameter
        `plotter`. Then by setting `return_plotter` to `True`, the function
        adds the cells to the plotter and returns it for further customization.

        Parameters
        ----------
        deepcopy: bool, Optional
            If True, a deep copy is created. Default is False.
        jupyter_backend: str, Optional
            The backend to use when plotting in a Jupyter enviroment.
            Default is 'pythreejs'.
        show_edges: bool, Optional
            If True, the edges of the mesh are shown as a wireframe.
            Default is True.
        notebook: bool, Optional
            If True and in a Jupyter enviroment, the plot is embedded
            into the Notebook. Default is False.
        theme: str, Optional
            The theme to use with PyVista. Default is None.
        scalars: Union[str, numpy.ndarray]
            A string that refers to a field in the celldata objects
            of the block of the mesh, or a NumPy array with values for
            each point in the mesh.
        window_size: tuple, Optional
            The size of the window, only is `notebook` is `False`.
            Default is None.
        return_plotter: bool, Optional
            If True, an instance of :class:`pyvista.Plotter` is returned
            without being shown. Default is False.
        config_key: tuple, Optional
            A tuple of strings that refer to a configuration for PyVista.
        plotter: pyvista.Plotter, Optional
            A plotter to use. If not provided, a plotter is created in the
            background. Default is None.
        cmap: Union[str, Iterable], Optional
            A color map for plotting. See PyVista's docs for the details.
            Default is None.
        camera_position: tuple, Optional
            Camera position. See PyVista's docs for the details. Default is None.
        lighting: bool, Optional
            Whether to use lighting or not. Default is None.
        edge_color: str, Optional
            The color of the edges if `show_edges` is `True`. Default is None,
            which equals to the default PyVista setting.
        return_img: bool, Optional
            If True, a screenshot is returned as an image. Default is False.
        show_scalar_bar: Union[bool, None], Optional
            Whether to show the scalar bar or not. A `None` value means that the option
            is governed by the configurations of the blocks. If a boolean is provided here,
            it overrides the configurations of the blocks. Default is None.
        **kwargs
            Extra keyword arguments passed to `pyvista.Plotter`, it the plotter
            has to be created.

        Returns
        -------
        Union[None, pv.Plotter, numpy.ndarray]
            A PyVista plotter if `return_plotter` is `True`, a NumPy array if
            `return_img` is `True`, or nothing.

        Example
        -------
        .. plot::
            :include-source: True

            from sigmaepsilon.mesh.plotting import pvplot
            from sigmaepsilon.mesh.downloads import download_gt40
            import matplotlib.pyplot as plt
            mesh = download_gt40(read=True)
            img=pvplot(mesh, notebook=False, return_img=True)
            plt.imshow(img)
            plt.axis('off')
        """
        if not isinstance(obj, PolyData):  # pragma: no cover
            raise TypeError(f"Expected PolyData, got {type(obj)}.")

        polys = obj.to_pv(deepcopy=deepcopy, multiblock=False, scalars=scalars)

        if isinstance(theme, str):
            try:
                new_theme_type = pv.themes._ALLOWED_THEMES[theme].value
                theme = new_theme_type()
            except Exception:
                if theme == "dark":
                    theme = themes.DarkTheme()
                    theme.lighting = False
                elif theme == "bw":
                    theme = themes.Theme()
                    theme.color = "black"
                    theme.lighting = True
                    theme.edge_color = "white"
                    theme.background = "white"
                elif theme == "document":
                    theme = themes.DocumentTheme()

        if theme is None:
            theme = pv.global_theme

        theme.show_edges = show_edges

        if lighting is not None:
            theme.lighting = lighting

        if edge_color is not None:
            theme.edge_color = edge_color

        if plotter is None:
            pvparams = dict()

            if window_size is not None:
                pvparams.update(window_size=window_size)

            pvparams.update(kwargs)
            pvparams.update(notebook=notebook)
            pvparams.update(theme=theme)

            if "title" not in pvparams:
                pvparams["title"] = "SigmaEpsilon"

            if not notebook and return_img:
                pvparams["off_screen"] = True

            plotter = pv.Plotter(**pvparams)

        if camera_position is not None:
            plotter.camera_position = camera_position

        pvparams = dict()
        blocks: Iterable[PolyData] = obj.cellblocks(inclusive=True, deep=True)

        blocks_have_data = obj._has_plot_scalars_(scalars)

        if config_key is None:
            config_key = obj.__class__._pv_config_key_

        for block, poly, has_data in zip(blocks, polys, blocks_have_data):
            NDIM = block.cd.Geometry.number_of_spatial_dimensions
            params = copy(pvparams)
            config = block._get_config_(config_key)
            if has_data:
                config.pop("color", None)
            params.update(config)

            if cmap is not None:
                params["cmap"] = cmap

            if NDIM > 1:
                params["show_edges"] = show_edges

            if isinstance(show_scalar_bar, bool):
                params["show_scalar_bar"] = show_scalar_bar

            plotter.add_mesh(poly, **params)

        if return_plotter:
            return plotter

        show_params = dict()
        if notebook:
            show_params.update(jupyter_backend=jupyter_backend)
        else:
            if return_img:
                plotter.show(auto_close=False)
                plotter.show(screenshot=True)
                return plotter.last_image

        return plotter.show(**show_params)


plotters["PyVista"] = pvplot

__all__ = ["pvplot"]



from ..config import __hask3d__
from ..helpers import plotters

if not __hask3d__:  # pragma: no cover

    def k3dplot(*_, **__) -> None:
        raise ImportError(
            "You need K3D for this. Install it with 'pip install k3d'. "
            "You may also need to restart your kernel and reload the package."
        )

else:
    from typing import Union, Optional

    import k3d

    from ..data import PolyData

    def k3dplot(
        obj: PolyData,
        scene: Optional[Union[k3d.Plot, None]] = None,
        *,
        menu_visibility: Optional[bool] = True,
        **kwargs,
    ) -> k3d.Plot:
        """
        Plots the mesh using 'k3d' as the backend.

        .. warning::
            During this call there is a UserWarning saying 'Given trait value dtype
            "float32" does not match required type "float32"'. Although this is weird,
            plotting seems to be just fine.

        Parameters
        ----------
        scene: object, Optional
            A K3D plot widget to append to. This can also be given as the
            first positional argument. Default is None, in which case it is
            created using a call to :func:`k3d.plot`.
        menu_visibility: bool, Optional
            Whether to show the menu or not. Default is True.
        **kwargs: dict, Optional
            Extra keyword arguments forwarded to :func:`to_k3d`.

        Example
        -------
        Get a compound mesh, add some random data to it and plot it with K3D.

        .. code-block:: python

            # doctest: +SKIP

            from sigmaepsilon.mesh.plotting import k3dplot
            from sigmaepsilon.mesh.examples import compound_mesh
            from k3d.colormaps import matplotlib_color_maps
            import k3d
            import numpy as np
            mesh = compound_mesh()
            cd_L2 = mesh["lines", "L2"].cd
            cd_Q4 = mesh["surfaces", "Q4"].cd
            cd_H8 = mesh["bodies", "H8"].cd
            cd_L2.db["scalars"] = 100 * np.random.rand(len(cd_L2))
            cd_Q4.db["scalars"] = 100 * np.random.rand(len(cd_Q4))
            cd_H8.db["scalars"] = 100 * np.random.rand(len(cd_H8))
            scalars = mesh.pd.pull("scalars")
            cmap = matplotlib_color_maps.Jet
            fig = k3d.plot()
            k3dplot(mesh, fig, scalars=scalars, menu_visibility=False, cmap=cmap)
            fig
        """
        if scene is None:
            scene = k3d.plot(menu_visibility=menu_visibility)
        return obj.to_k3d(scene=scene, **kwargs)


plotters["k3d"] = k3dplot

__all__ = ["k3dplot"]



# -*- coding: utf-8 -*-
from .l2 import L2
from .l2 import L2 as Line
from .l3 import L3
from .l3 import L3 as QuadraticLine
from .t3 import T3
from .t3 import T3 as Tri
from .q4 import Q4
from .q4 import Q4 as Quad
from .q8 import Q8
from .q9 import Q9
from .t6 import T6
from .h8 import H8
from .h8 import H8 as Hex
from .h27 import H27
from .tet4 import TET4
from .tet4 import TET4 as Tetra
from .tet10 import TET10
from .w6 import W6
from .w6 import W6 as Wedge
from .w18 import W18

__all__ = [
    "L2",
    "Line",
    "L3",
    "QuadraticLine",
    "T3",
    "Tri",
    "Q4",
    "Quad",
    "Q8",
    "Q9",
    "T6",
    "H8",
    "Hex",
    "H27",
    "TET4",
    "Tetra",
    "TET10",
    "W6",
    "Wedge",
    "W18",
]



from sigmaepsilon.mesh.space import CartesianFrame
from sigmaepsilon.mesh.recipes import cylinder
from sigmaepsilon.mesh import PolyData, LineData, PointData
from sigmaepsilon.mesh.cells import H8, Q4, L2
from sigmaepsilon.mesh.utils.topology import H8_to_L2, H8_to_Q4
from sigmaepsilon.mesh.utils.topology import detach_mesh_bulk
from sigmaepsilon.mesh.utils.space import frames_of_lines, frames_of_surfaces
from sigmaepsilon.math import minmax
import numpy as np


def compound_mesh() -> PolyData:
    """
    Returns a compound mesh with H8, Q4 and L2 cells in it.
    """
    min_radius = 5
    max_radius = 25
    h = 50

    shape = (min_radius, max_radius), h
    frame = CartesianFrame(dim=3)
    cyl = cylinder(shape, size=5.0, voxelize=True, frame=frame)

    coords = cyl.coords()
    topo = cyl.topology().to_numpy()
    centers = cyl.centers()

    cxmin, cxmax = minmax(centers[:, 0])
    czmin, czmax = minmax(centers[:, 2])
    cxavg = (cxmin + cxmax) / 2
    czavg = (czmin + czmax) / 2
    b_upper = centers[:, 2] > czavg
    b_lower = centers[:, 2] <= czavg
    b_left = centers[:, 0] < cxavg
    b_right = centers[:, 0] >= cxavg
    iL2 = np.where(b_upper & b_right)[0]
    iTET4 = np.where(b_upper & b_left)[0]
    iH8 = np.where(b_lower)[0]
    _, tL2 = H8_to_L2(coords, topo[iL2])
    _, tQ4 = H8_to_Q4(coords, topo[iTET4])
    tH8 = topo[iH8]

    pd = PointData(coords=coords, frame=frame)
    mesh = PolyData(pd)

    cdL2 = L2(topo=tL2, frames=frames_of_lines(coords, tL2))
    mesh["lines", "L2"] = LineData(cdL2)

    cdQ4 = Q4(topo=tQ4, frames=frames_of_surfaces(coords, tQ4))
    mesh["surfaces", "Q4"] = PolyData(cdQ4)

    cH8, tH8 = detach_mesh_bulk(coords, tH8)
    pdH8 = PointData(coords=cH8, frame=frame)
    cdH8 = H8(topo=tH8, frames=frame)
    mesh["bodies", "H8"] = PolyData(pdH8, cdH8)

    mesh.to_standard_form()
    mesh.lock(create_mappers=True)

    return mesh



from numba import njit, prange
import numpy as np
from numpy import ndarray

__cache = True


@njit(nogil=True, cache=__cache)
def monoms_W6_single(x: ndarray) -> ndarray:
    r, s, t = x
    return np.array([1, r, s, t, r * t, s * t], dtype=x.dtype)


@njit(nogil=True, parallel=True, cache=__cache)
def monoms_W6_multi(x: ndarray) -> ndarray:
    nP = x.shape[0]
    res = np.zeros((nP, 6), dtype=x.dtype)
    for i in prange(nP):
        res[i] = monoms_W6_single(x[i])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def monoms_W6_bulk_multi(x: ndarray) -> ndarray:
    nE = x.shape[0]
    res = np.zeros((nE, 6, 6), dtype=x.dtype)
    for i in prange(nE):
        res[i] = monoms_W6_multi(x[i])
    return res


def monoms_W6(x: ndarray) -> ndarray:
    N = len(x.shape)
    if N == 1:
        return monoms_W6_single(x)
    elif N == 2:
        return monoms_W6_multi(x)
    elif N == 3:
        return monoms_W6_bulk_multi(x)
    else:
        raise NotImplementedError



from typing import (
    Union,
    Callable,
    Optional,
    ClassVar,
    Iterable,
    Protocol,
    runtime_checkable,
)

from numpy import ndarray


__all__ = ["GeometryProtocol"]


@runtime_checkable
class GeometryProtocol(Protocol):
    """
    Protocol for Geometry classes.
    """

    number_of_nodes: ClassVar[int]
    number_of_spatial_dimensions: ClassVar[int]
    number_of_nodal_variables: ClassVar[int] = 1
    vtk_cell_id: ClassVar[Optional[int]] = None
    meshio_cell_id: ClassVar[Optional[str]] = None
    boundary_class: ClassVar[Optional["GeometryProtocol"]] = None
    shape_function_evaluator: ClassVar[Optional[Callable]] = None
    shape_function_matrix_evaluator: ClassVar[Optional[Callable]] = None
    shape_function_derivative_evaluator: ClassVar[Optional[Callable]] = None
    monomial_evaluator: ClassVar[Optional[Callable]] = None
    quadrature: ClassVar[Optional[dict]] = None

    @classmethod
    def master_coordinates(cls) -> ndarray:
        """
        Returns the coordinates of the master element.

        Returns
        -------
        numpy.ndarray
        """
        ...

    @classmethod
    def master_center(cls) -> ndarray:
        """
        Returns the coordinates of the master element.

        Returns
        -------
        numpy.ndarray
        """
        ...

    @classmethod
    def shape_function_values(
        cls, x: Union[float, Iterable[float]], *args, **kwargs
    ) -> ndarray:
        """
        Evaluates the shape functions at the specified locations.
        """
        ...

    @classmethod
    def shape_function_derivatives(
        cls, x: Union[float, Iterable[float]], *args, **kwargs
    ) -> ndarray:
        """
        Evaluates shape function derivatives wrt. the master element or the local
        coordinate frames of some cells. To control the behaviour, either 'jac' or 'wrt'
        can be provided.
        """
        ...

    @classmethod
    def shape_function_matrix(
        cls, x: Union[float, Iterable[float]], *args, N: Optional[int] = 1, **kwargs
    ) -> ndarray:
        """
        Evaluates the shape function matrix at the specified locations and
        an N number of nodal variables.
        """
        ...

    @classmethod
    def approximator(cls, *args, **kwargs) -> Callable:
        """
        Returns a callable object that can be used to approximate over
        nodal values of one or more cells.
        """
        ...



from typing import Tuple, List

import numpy as np
from numpy import ndarray
from sympy import symbols

from ..geometry import PolyCellGeometry3d
from ..data.polycell import PolyCell
from ..utils.cells.tet10 import (
    monoms_TET10,
)
from ..utils.numint import Gauss_Legendre_Tet_4


class TET10(PolyCell):
    """
    10-node isoparametric hexahedron.
    """

    label = "TET10"

    class Geometry(PolyCellGeometry3d):
        number_of_nodes = 10
        vtk_cell_id = 24
        monomial_evaluator = monoms_TET10
        quadrature = {
            "full": Gauss_Legendre_Tet_4,
            "geometry": "full",
        }

        @classmethod
        def polybase(cls) -> Tuple[List]:
            """
            Retruns the polynomial base of the master element.

            Returns
            -------
            list
                A list of SymPy symbols.
            list
                A list of monomials.
            """
            locvars = r, s, t = symbols("r s t", real=True)
            monoms = [1, r, s, t, r * s, r * t, s * t, r**2, s**2, t**2]
            return locvars, monoms

        @classmethod
        def master_coordinates(cls) -> ndarray:
            return np.array(
                [
                    [-1 / 3, -1 / 3, -1 / 3],
                    [2 / 3, -1 / 3, -1 / 3],
                    [-1 / 3, 2 / 3, -1 / 3],
                    [-1 / 3, -1 / 3, 2 / 3],
                    [1 / 6, -1 / 3, -1 / 3],
                    [1 / 6, 1 / 6, -1 / 3],
                    [-1 / 3, 1 / 6, -1 / 3],
                    [-1 / 3, -1 / 3, 1 / 6],
                    [1 / 6, -1 / 3, 1 / 6],
                    [-1 / 3, 1 / 6, 1 / 6],
                ]
            )

        @classmethod
        def master_center(cls) -> ndarray:
            return np.array([[0.0, 0.0, 0.0]], dtype=float)

        @classmethod
        def tetmap(cls, subdivide: bool = True) -> np.ndarray:
            if subdivide:
                raise NotImplementedError
            else:
                return np.array([[0, 1, 2, 3]], dtype=int)



from .pvplot import pvplot
from .k3dplot import k3dplot
from .mpl import (
    triplot_mpl_hinton,
    triplot_mpl_mesh,
    triplot_mpl_data,
    parallel_mpl,
    aligned_parallel_mpl,
    decorate_mpl_ax,
)
from .plotly import (
    scatter_lines_plotly,
    scatter_points_plotly,
    plot_lines_plotly,
    triplot_plotly,
)

__all__ = [
    "pvplot",
    "triplot_mpl_hinton",
    "triplot_mpl_mesh",
    "triplot_mpl_data",
    "parallel_mpl",
    "aligned_parallel_mpl",
    "decorate_mpl_ax",
    "scatter_lines_plotly",
    "scatter_points_plotly",
    "plot_lines_plotly",
    "triplot_plotly",
    "k3dplot",
]
